/*
 * j_conv.c
 *
 * By Ross Ridge
 * Public Domain
 *
 */

#include "defs.h"

#include <limits.h>

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) myjis j_conv.c 1.7 98/06/01 22:28:04";
#endif

struct jis_code_name const __jis_code_name_table[] = {
	{ "iso8859-1",	JIS_LOCALE_ASCII },
	{ "ascii",	JIS_LOCALE_ASCII },
	{ "iso646",	JIS_LOCALE_ASCII },
	{ "jis-roman",	JIS_LOCALE_JIS_ROMAN },
	{ "jis201",	JIS_LOCALE_JIS_ROMAN },
	{ "jis",	JIS_LOCALE_JIS },
	{ "new-jis",	JIS_LOCALE_JIS },
	{ "jis208",	JIS_LOCALE_JIS },
	{ "euc-jp",	JIS_LOCALE_EUC },
	{ "euc",	JIS_LOCALE_EUC },
	{ "iso-2022-jp", JIS_LOCALE_EUC },
	{ "shiftjis",	JIS_LOCALE_SHIFT_JIS },
	{ "sjis",	JIS_LOCALE_SHIFT_JIS },
#ifdef __JIS_CONFIG_HAVE_UNICODE
	{ "utf-8",	JIS_LOCALE_UTF8 },
	{ "utf8",	JIS_LOCALE_UTF8 },
#endif
	{ "wide_char",	JIS_LOCALE_DEFAULT },
	{ NULL, 0 },
};

jis_conv_state
jis_conv_open(char const *to, char const *from) {
	int i;
	jis_conv_state p;
	struct jis_code_name const * const table = __jis_code_name_table;
	enum jis_locale_ctypes inloc, outloc;

	i = -1;
	do {
		i++;
		if (table[i].name == NULL) {
			errno = EINVAL;
			return NULL;
		}
	} while (strcmp(from, table[i].name) != 0);
	inloc = table[i].loc;

	i = -1;
	do {
		i++;
		if (table[i].name == NULL) {
			errno = EINVAL;
			return NULL;
		}
	} while (strcmp(to, table[i].name) != 0);
	outloc = table[i].loc;

	p = malloc(sizeof(struct __jis_conv_state));
	if (p == NULL) {
		return p;
	}

	p->ostate = p->istate = __jis_initial_state;

	if (inloc == JIS_LOCALE_DEFAULT) {
		p->widein = 1;
	} else {
		p->widein = 0;
		jis_set_state_locale(&p->istate, inloc);
	}

	if (outloc == JIS_LOCALE_DEFAULT) {
		p->wideout = 1;
	} else {
		p->wideout = 0;
		jis_set_state_locale(&p->ostate, outloc);
	}

	return p;
}

size_t
jis_conv(jis_conv_state statep, char const **inp, size_t *ilenp,
	 char **outp, size_t *olenp) {
	wide_char wc;
	int i;
	char const *in = *inp;
	char *out = *outp;
	size_t ilen = *ilenp;
	size_t olen = *olenp;
	jis_state istate = statep->istate;
	jis_state ostate = statep->ostate;
	int widein = statep->widein;
	int wideout = statep->wideout;
	size_t ret = 0;
	size_t iclen, oclen;
	unsigned char tmp;
	
	while(ilen > 0) {
		if (widein) {
			if (ilen < sizeof(wide_char)) {
				errno = EINVAL;
				break;
			}

			tmp = (unsigned char) in[sizeof(wide_char) - 1];
			for (i = sizeof(wide_char) - 2; i >= 0; i--) {
				tmp *= UCHAR_MAX + 1;
				tmp += (unsigned char) in[i];
			}
			wc = (wide_char) tmp;
			iclen = 4;
		} else {
			iclen = jis_to_wide(&wc, in, ilen, &istate);
			if (iclen == (size_t) -1) {
				break;
			}
			if (iclen == (size_t) -2) {
				/*
				 * The entire input buffer has been
				 * read and used to update the state.
				 */
				statep->istate = istate;
				in += ilen;
				ilen = 0;
				break;
			}
		}
		if (wideout) {
			if (olen < sizeof(wide_char)) {
				errno = E2BIG;
				break;
			}
			tmp = (unsigned long) wc;
			for (i = 0; i < (int) sizeof(wide_char); i++) {
				out[i] = (unsigned char) wc;
				wc /= UCHAR_MAX + 1;
			}
			oclen = sizeof(wide_char);
		} else {
			oclen = wide_to_jis(out, olen, wc, &ostate);
			if (oclen == (size_t) -1) {
#if 0
				if (errno == EILSEQ) {
					fprintf(stderr,
						"@@@ %08lx\n", (long) wc);
				}
#endif
				if (errno == EILSEQ) {
					ostate = statep->ostate;
					wc = (wide_char) '?';
					oclen = wide_to_jis(out, olen, wc,
							    &ostate);
					assert(oclen != (size_t) -1);
				} else {
					break;
				}
			} else if (oclen == 0) {
				errno = E2BIG;
				break;
			}
		}

		assert(ilen >= iclen && olen >= oclen);
		ilen -= iclen;
		in += iclen;
		olen -= oclen;
		out += oclen;
		statep->istate = istate;
		statep->ostate = ostate;

		ret++;
	}
	
	*ilenp = ilen;
	*olenp = olen;
	*inp = in;
	*outp = out;

	if (ilen != 0) {
		return (size_t) -1;
	}
	
	return ret;
}
	
int
jis_conv_close(jis_conv_state p) {
	free(p);
	return 0;
}
	




