/*
 * kd_wrind.c
 *
 * By Ross Ridge
 * Public Domain
 *
 */

#include "defs.h"
#include "kanjidic.h"

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) myjis kd_rdind.c 1.4 98/06/01 22:28:20";
#endif


static int
read_32(FILE *f, long *val) {
	unsigned char n[4];
	int i;
	for(i = 0; i < 4; i++) {
		int c = getc(f);
		if (c == EOF) {
			printf("@@@ error in getc 32\n");
			return -1;
		}
		n[i] = c;
	}

	if (n[3] > 127) {
		printf("@@@ num to big 32\n");
		errno = EINVAL;
		return -1;
	}

	*val = (n[0] & 0xFFUL) + ((n[1] & 0xFFUL) << 8) 
		+ ((n[2] & 0xFFUL) << 16) + ((n[3] & 0xFFUL) << 24);

	return 0;
}

static int
read_16(FILE *f, unsigned *val) {
	unsigned char n[2];
	int i;
	for(i = 0; i < 2; i++) {
		int c = getc(f);
		if (c == EOF) {
			printf("@@@ error in getc 16\n");
			return -1;
		}
		n[i] = c;
	}

	*val = (n[0] & 0xFFU) | ((n[1] & 0xFFU) << 8);

	return 0;
}


inline int
read_wchar(FILE *f, wide_char *wc) {
#ifdef __JIS_CONFIG_USE_UNICODE
	long n;
	if (read_32(f, &n) == -1) {
		return -1;
	}
#else
	unsigned n;
	if (read_16(f, &n) == -1) {
		return -1;
	}
#endif
	*wc = (wide_char) n;

	return 0;
}

template <class TYPE>
int
read_ent(FILE *f, list<TYPE> *lst) {
	int len;
	if (read_ent(f, &len) == -1) {
		return -1;
	}
	*lst = NULL;
	if (lst->extend(len) != 0) {
		printf("@@@ error extending\n");
		return -1;
	}
	int i;
	for (i = 0; i < len; i++) {
		if (read_ent(f, &lst->select(i)) == -1) {
			printf("@@@ error reading list elt. %d\n", sizeof(TYPE));
			return -1;
		}
	}
	return 0;
}

static inline int
read_ent(FILE *f, int *n) {
	unsigned u;
	if (read_16(f, &u) == -1) {
		return -1;
	}
	if (u > 32767) {
		errno = EINVAL;
		return -1;
	}
	*n = (int) u;
	return 0;
}

template <class TYPE>
static int
read_ent(FILE *f, element<TYPE> *elt) {
	if (read_ent(f, &elt->key) == -1) {
		return -1;
	}
	int n;
	if (read_ent(f, &n) == -1) {
		return -1;
	}
	elt->line = (short) n;
	return 0;
}

template <class TYPE>
static int
read_ent(FILE *f, index<TYPE> *lst) {
	int r = read_ent(f, (list<element<TYPE>/**/> *) lst);
	if (r == -1) {
		printf("@@@ error reading index\n");
	}
	return r;
}

static inline int
read_ent(FILE *f, skip_code *sc) {
	unsigned u;
	if (read_16(f, &u) == -1) {
		return -1;
	}
	*sc = skip_code((short unsigned) u);
	return 0;
}

static inline int
read_ent(FILE *f, wide_char **ws) {
	unsigned len;
	if (read_16(f, &len) == -1) {
		return -1;
	}

	wide_char *s = (wide_char *) malloc((len + 1) * sizeof(wide_char));
	if (s == NULL) {
		return -1;
	}

	int i;
	for(i = 0; i < (int) len; i++) {
		if (read_wchar(f, s + i) == -1) {
			free(s);
			return -1;
		}
	}
	s[len] = 0;

	*ws = s;

	return 0;
}

static inline int
read_ent(FILE *f, struct kanjidict::quick_entry *qent) {
	unsigned score;
	if (read_32(f, &qent->offset) == -1
	    || read_wchar(f, &qent->character) == -1
	    || read_16(f, &score) == -1) {
		printf("@@@ error reading quick_ent\n");
		return -1;
	}
	qent->score = (unsigned short) score;
	return 0;
}

static int
read_radical_index(FILE *f, list<int> *lsts) {
	int i;
	for (i = 0; i < 214; i++) {
		if (read_ent(f, lsts + i) == -1) {
			printf("@@@ error in radical index\n");
			return -1;
		}
	}
	return 0;
}

int
kanjidict::read_index(char const *ifile) {
	FILE *f = fopen(ifile, "rb");
	if (f == NULL) {
		return -1;
	}

	if (read_ent(f, &quick_table) == -1
	    || read_ent(f, &skip_index) == -1
	    || read_ent(f, &stroke_index) == -1
	    || read_radical_index(f, radical_index) == -1
	    || read_ent(f, static_cast<index<reading> *>(&reading_index)) == -1
	    || read_ent(f, static_cast<index<reading> *>(&name_reading_index))
	       == -1) {
		printf("@@@ error reading file\n");
		return -1;
	}

	if (feof(f)) {
		errno = EIO;
		return -1;
	}

	if (fclose(f) == EOF) {
		return -1;
	}

	return 0;
}

/* Local variables: */
/* mode: c++ */
/* End: */


