/*
 * kl.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * @(#) myjis kl.c 1.4 00/05/11 18:17:14
 *
 */

#include "defs.h"
#include "edict.h"
#include <ctype.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef CONFIG_SCCS_IDS
static char const sccs_id[] = "@(#) myjis kl.c 1.4 00/05/11 18:17:14";
#endif

static void *
icheck(void *p) {
	if (p == NULL && errno == ENOMEM) {
		xcheck(p);
	}
	return p;
}

static void
write_error(void) {
	perror("stdout");
	exit(1);
}

static char *
name_to_index(char const *name) {
	char *s = xmalloc(strlen(name) + 6 + 1);
	strcpy(s, name);
	return strcat(s, ".index");
}

static wide_char *
kana_to_roman_dup(wide_char const *ws, int flags, int *r) {
        wide_char *kana;
	wide_char *buf;
	size_t len = wide_len(ws) * 3 + 1;

	kana = xcheck(wide_dup(ws));
	wide_to_hiragana_vowel_str(kana);

	while(1) {
		size_t tmplen = len;
		wide_char const *tmpkana = kana;
	
		buf = xmalloc(len * sizeof(wide_char)); 
		*r = wide_kana_to_roman_str(buf, &tmplen, &tmpkana, flags);
		if (*tmpkana == '\0') {
			break;
		}
		free(buf);
		len *= 2;
	}

	free(kana);

	return buf;
}

static int
romanize(char const *line, int rflags, int banner) {
	wide_char *wline;
	wide_char *word1;
	wide_char *word2 = NULL;
	wide_char *defn = NULL;
	wide_char *ws;
	int r;
	int ret = 0;
	char *s;

	wline = icheck(jis_to_wide_dup(line));
	if (wline == NULL) {
		return -1;
	}

	ws = wide_chr(wline, (wide_char) '\n');
	if (ws != NULL) {
		*ws = '\0';
	}

	word1 = wline;
	ws = wide_chr(wline, (wide_char) ' ');
	if (ws != NULL) {
		*ws++ = '\0';
		if (*ws == '[') {
			word2 = ws;
			ws = wide_chr(ws + 1, (wide_char) ' ');
			if (ws != NULL) {
				*ws++ = '\0';
				defn = ws;
			}
		} else {
			defn = ws;
		}
	}

	ws = kana_to_roman_dup(word1, rflags | JIS_R2K_EXTENDED, &r);
	if (!r) {
		word1 = ws;
	} else if (!banner) {
		word1 = NULL;
	}

	if (word1 != NULL) {
		s = wide_to_jis_dup(word1);
		if (s == NULL) {
			free(ws);
			free(wline);
			return -1;
		}

		if (r) {
			FILE *f;

			fflush(stdout);
			f = popen("kbanner -s -5 | head -8", "w");
			if (f == NULL) {
				ret = 1;
			} else {
				fprintf(f, "%s\n", s);
				if (pclose(f) != 0) {
					ret = 1;
				}
			}
		} else {
			if (printf("%s", s) < 0) {
				write_error();
			}
			if (word2 != NULL || defn != NULL) {
				if (putchar(' ') == EOF) {
					write_error();
				}
			}
		}
		free(s);
	}

	free(ws);

	if (word2 != NULL) {
		ws = kana_to_roman_dup(word2, rflags, &r);
		s = icheck(wide_to_jis_dup(ws));
		if (s == NULL) {
			free(ws);
			free(wline);
			return -1;
		}
		if (printf("%s", s) < 0) {
			write_error();
		}
		if (defn != NULL) {
			if (putchar(' ') == EOF) {
				write_error();
			}
		}
		free(s);
		free(ws);
	}


	if (defn != NULL) {
		s = icheck(wide_to_jis_dup(defn));
		if (s == NULL) {
			free(wline);
			return -1;
		}
		if (printf("%s", s) == EOF) {
			write_error();
		}
		free(s);
	}

	if (putchar('\n') == EOF) {
		write_error();
	}

	free(wline);

	return ret;
}

static int
prepare_dictionary(char const *name) {
	struct stat sb;
	static char const dict_dir[] = CONFIG_DICTIONARY_DIR;
	static char const comp_dir[] = CONFIG_COMPRESSED_DICTIONARY_DIR;
	int fd_in, fd_out;
	pid_t pid;
	int st;
	mode_t user_mask = umask(0);
	int force_index = 0;
	char *comp_name;
	char *index_name;

	if (mkdir(dict_dir, 0777) == -1 && errno != EEXIST) {
		perror(dict_dir);
		umask(user_mask);
		return -1;
	}

	umask(user_mask);

	if (chdir(dict_dir) == -1) {
		perror(dict_dir);
		return -1;
	}

	if (stat(name, &sb) == -1) {
		if (errno != ENOENT) {
			perror(name);
			return -1;
		}

		comp_name = xmalloc(strlen(comp_dir) 
				    + 1 + strlen(name) + 3 + 1);
		strcpy(comp_name, comp_dir);
		strcat(comp_name, "/");
		strcat(comp_name, name);
		strcat(comp_name, ".gz");

		fd_in = open(comp_name, O_RDONLY);
		if (fd_in == -1) {
			perror(comp_name);
			free(comp_name);
			return -1;
		}
		free(comp_name);

		fprintf(stderr, "Uncompressing %s to %s/%s... ", 
			comp_name, dict_dir, name);

		umask(user_mask & (mode_t) ~0444);
		fd_out = open(name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
		if (fd_out == -1) {
			putc('\n', stderr);
			perror(name);
			umask(user_mask);
			close(fd_in);
			return -1;
		}
		umask(user_mask);

		fflush(stdout);
		fflush(stderr);

		pid = fork();
		if (pid == (pid_t) -1) {
			perror("\nfork");
			close(fd_in);
			close(fd_out);
			return -1;
		} else if (pid == 0) {
			close(0);
			dup(fd_in);
			close(fd_in);
			close(1);
			dup(fd_out);
			close(fd_out);
			execlp("gunzip", "gunzip", "-c", "-", NULL);
			perror(name);
			_exit(127);
		}
		
		close(fd_in);
		close(fd_out);
		if (wait(&st) == -1) {
			perror("\nwait");
			unlink(name);
			return -1;
		}

		putc('\n', stderr);

		fflush(stderr);

		if (WIFEXITED(st) && WEXITSTATUS(st) != 0) {
			fprintf(stderr, "gunzip exited with status %d\n",
				WEXITSTATUS(st));
			unlink(name);
			return -1;
		} else if (WIFSIGNALED(st)) {
			fprintf(stderr, "gunzip died on signal %d\n",
				WTERMSIG(st));
			unlink(name);
			return -1;
		}
		force_index = 1;
	}			       

	index_name = name_to_index(name);

	if (force_index || stat(index_name, &sb) == -1) {
		if (errno != ENOENT) {
			perror(index_name);
			free(index_name);
			return -1;
		}
		fprintf(stderr, "Indexing %s/%s...", dict_dir, name);
		if (edict_make_index(index_name, name, 1) == -1) {
			putc('\n', stderr);
			perror(name);
			free(index_name);
			return -1;
		}	
		if (stat(index_name, &sb) == 0) {
			chmod(index_name, sb.st_mode | 0444);
		}
	}
	free(index_name);

	return 0;
}


static int
lookup3(struct edict *dict, char const *key, int exact_match, int ending_match,
	int roman, int rflags, int banner) {

	int ret = 0;
	int err;

	if (edict_lookup(dict, key) == -1) {
		fprintf(stderr, "corrupted dictionary\n");
		return -1;
	}
	while(1) {
		char *entry = edict_get_entry(dict, exact_match, ending_match);
		if (entry == NULL) {
			if (errno == ENOENT) {
				return 0;
			}
			fprintf(stderr, "corrupted dictionary\n");
			return -1;
		}
		if (ret == 0) {
			ret = 1;
		}
		if (roman) {
			if (romanize(entry, rflags, banner) == -1) {
				err = errno;
				ret = -1;
			}
		} else if (fputs(entry, stdout) == EOF) {
			write_error();
		}
	}

	if (ret == -1) {
		errno = err;
	}

	return ret;
}

static int
lookup2(struct edict *edict, char *word, int reduce,
	int exact_match, int roman, int rflags, int banner) {
	int ret;
	int r;

	if (reduce) {
		char **words;
		char **p;

		ret = lookup3(edict, word, 1, 0, roman, rflags, banner);
		words = icheck(edict_reduce(word));
		if (words == NULL) {
			perror("reduce");
			return -1;
		}
		p = words;
		while(1) {
			if (*p == NULL) {
				break;
			}
			r = lookup3(edict, *p, 1, 1, roman, rflags, banner);
			if (r == -1 || ret == 0) {
				ret = r;
			}
			free(*p);
			p++;
		}
		free(words);
		return ret;
	} 
	return lookup3(edict, word, exact_match, 0, roman, rflags, banner);
}

static int
hasletter(char const *s) {
	size_t r;
	jis_state state = __jis_initial_state;

	while(!isalpha((unsigned char) *s)) {
		if (*s == '\0') {
			return 0;
		}
		r = jis_to_wide(NULL, s, INT_MAX, &state);
		if (r == 0 || r == (size_t) -1 || r == (size_t) -2) {
			return 0;
		}
		s += r;
	}

	return 1;
}

static int
lookup(struct edict *edict, char *word, int reduce, int allkana,
       int exact_match, int roman, int rflags, int banner) {
	size_t klen;
	size_t i;
	int r;
	
	if (hasletter(word)) {
		int err;
		int ret = 0;
		wide_char *kwords[CONFIG_MAX_AMBIGUOUS];
		wide_char *wword = icheck(jis_to_wide_dup(word));

		if (wword == NULL) {
			perror(word);
			return -1;
		}

		if (allkana) {
			klen = CONFIG_MAX_AMBIGUOUS;
		} else {
			klen = 1;
		}

		klen = wide_roman_to_kana_list(kwords, klen, wword,
					       rflags, &err);

		free(wword);
		if ((err & JIS_R2K_ERR_AMBIGUOUS) && allkana) {
			fprintf(stderr, "%s: too ambiguous\n", word);
			ret = -1;
		}

		for(i = 0; i < klen; i++) {
			char *jword = icheck(wide_to_jis_dup(kwords[i]));
			if (jword == NULL) {
				perror(word);
			} else {
				r = lookup2(edict, jword, reduce, exact_match,
					    roman, rflags, banner);
				if (r == -1 || ret == 0) {
					ret = r;
				}
				free(jword);
			}
			free(kwords[i]);
		}
		return ret;
	}
	return lookup2(edict, word, reduce, exact_match,
		       roman, rflags, banner);
}

static int
icase_cmpn(wide_char const *ws1, wide_char const *ws2, size_t len) {
	size_t i;

	for (i = 0; i < len; i++) {
		wide_char wc1 = ws1[i];
		wide_char wc2 = ws2[i];
		char c1, c2;

		if (wc1 == wc2) {
			if (wc1 == (wide_char) 0) {
				return 0;
			}
			continue;
		}

		c1 = jis_wide_to_byte(wc1);
		c2 = jis_wide_to_byte(wc2);
		if (c1 != EOF && c2 != EOF && toupper(c1) == toupper(c2)) {
			continue;
		}
		return wc1 - wc2;
	}
	return 0;
}

static int
rlookup(struct edict *dict, char *word,	int exact_match,
	int roman, int rflags, int banner) {
	wide_char *wkey = icheck(jis_to_wide_dup(word));
	size_t wkey_len = wide_len(wkey);
	char *entry;
	wide_char *wentry;

	if (wkey == NULL) {
		perror(word);
		return -1;
	}

	edict_rewind(dict);
       
	entry = NULL;
	wentry = NULL;
	while(1) {
		wide_char *ws;

		free(entry);

		entry = edict_get_entry(dict, 0, 0);
		if (entry == NULL) {
			break;
		}

		free(wentry);
		wentry = icheck(jis_to_wide_dup(entry));
		if (wentry == NULL) {
			break;
		}

		ws = wide_chr(wentry, (wide_char) '/');
		if (ws == NULL) {
			errno = EINVAL;
			break;
		}

		if (exact_match) {
			if (icase_cmpn(wkey, ws + 1, wkey_len) != 0
			    || ws[1 + wkey_len] != '/') {
				continue;
			}
		} else {
			while(*ws != '\0') {
				ws++;
				if (icase_cmpn(wkey, ws, wkey_len) == 0) {
					break;
				}
				while(*ws != ' ' && *ws != '/' 
				      && *ws != '\0') {
					ws++;
				}
			}
			if (*ws == '\0') {
				continue;
			}
		}

		if (roman) {
			if (romanize(entry, rflags, banner) == -1) {
				break;
			}
		} else if (fputs(entry, stdout) == EOF) {
			write_error();
		}
	}

	free(entry);
	free(wentry);
	free(wkey);

	if (errno == ENOENT) {
		return 0;
	}
	if (errno == EINVAL) {
		fprintf(stderr, "corrupted dictionary\n");
	} else {
		perror("reading dictionary entry");
	}
	return -1;
}

static int
do_lookup(int dlen, struct edict **dicts, int wlen, char **words,
	  int reverse, int reduce, int allkana, int exact_match,
	  int roman, int rflags, int banner) {
	int ret = 0;
	int i;
	int j;
	int r;

	for(i = 0; i < wlen; i++) {
		for(j = 0; j < dlen; j++) {
			if (reverse) {
				r = rlookup(dicts[j], words[i], exact_match,
					    roman, rflags, banner);
			} else {
				r = lookup(dicts[j], words[i],
					   reduce, allkana, exact_match,
					   roman, rflags, banner);
			}
			if (r == -1) {
				ret = 2;
			} else if (ret == 0) {
				ret = r;
			}
		}
	}
	return ret;
}

static void
usage(char const *prgname) {
	fprintf(stderr, 
		"usage: %s [-rHKNjesucmRbn] words-to-look-up...\n"
		"\t-r	Reverse look up (English to Japanese)\n"
		"\t-m	Exact match\n"
		"\t-H	Use Hepburn romanization\n"
		"\t-K	Use Kunrei romanization\n"
		"\t-N	Use Nippon romanization\n"
		"\t-x	Use Extended romanization\n"
		"\t-j	Use JIS character encoding\n"
		"\t-e	Use EUC character encoding\n"
		"\t-s	Use Shift-JIS character encoding\n"
#ifdef __JIS_CONFIG_HAVE_UNICODE
		"\t-u	Use UTF-8 character encoding\n"
#else
		"\t-u	Use UTF-8 character encoding (not supported)\n"
#endif
		"\t-c	Use ASCII character encoding\n"
		"\t-a	Look up all possibilities for ambiguous"
		" romanizations\n"
		"\t-R	Reduce to dictionary form\n"
		"\t-b	Print kanji with kbanner\n"
		"\t-n	Search name dictionary as well\n",
		prgname);
	exit(2);
}

int
main(int argc, char **argv) {
	int name_dict = 0;
	int rflags = JIS_R2K_OBSOLETE | JIS_R2K_KATAKANA;
	int exact_match = 0;
	int reduce = 0;
	int banner = 0;
	int roman = 0;
	int allkana = 0;
	int reverse = 0;
	enum jis_locale_ctypes loc;
	int c;
	int dlen = 1;
	int i;
	char const *dict_names[2] = { CONFIG_MAIN_DICTIONARY };
	struct edict *dicts[2];


	setlocale(LC_ALL, "");
	loc = jis_set_locale_ctype(JIS_LOCALE_DEFAULT);

	while(1) {
		c = getopt(argc, argv, "rHKNjesucmRbnxa");
		if (c == EOF) {
			break;
		}
		switch(c) {
		case 'a':
			allkana = 1;
			break;

		case 'j':
			loc = JIS_LOCALE_JIS;
			break;

		case 's':
			loc = JIS_LOCALE_SHIFT_JIS;
			break;

		case 'e':
			loc = JIS_LOCALE_EUC;
			break;

#ifdef __JIS_CONFIG_HAVE_UNICODE
		case 'u':
			loc = JIS_LOCALE_UTF8;
			break;
#else /* __JIS_CONFIG_HAVE_UNICODE */
	        case 'u':
			fprintf(stderr, "%s: Sorry, UTF-8 encoding is not"
					" supported.", argv[0]);
			return 1;
#endif /* __JIS_CONFIG_HAVE_UNICODE */
		
		case 'c':
			loc = JIS_LOCALE_EUC;
			roman = 1;
			break;

		case 'H':
			rflags |= JIS_R2K_HEPBURN;
			break;

		case 'N':
			rflags |= JIS_R2K_NIPPON;
			break;
	
		case 'K':
			rflags |= JIS_R2K_KUNREI;
			break;

		case 'x':
			rflags |= JIS_R2K_EXTENDED;
			break;

		case 'm':
			exact_match = 1;
			break;

		case 'r':
			reverse = 1;
			break;

		case 'R':
			reduce = 1;
			break;

		case 'b':
			banner = 1;
			break;

		case 'n':
			name_dict = 1;
			break;

		case '?':
			usage(argv[0]);
		}
	}

	if (optind == argc) {
		fprintf(stderr, "no words to lookup\n");
		usage(argv[0]);
	}

	switch(loc) {
	case JIS_LOCALE_EUC:
	case JIS_LOCALE_SHIFT_JIS:
	case JIS_LOCALE_JIS:
#ifdef __JIS_CONFIG_HAVE_UNICODE
	case JIS_LOCALE_UTF8:
#endif
		break;

	default:
		loc = JIS_LOCALE_EUC;
		roman = 1;
	}

	jis_set_locale_ctype(loc);

	if (name_dict) {
		dict_names[dlen++] = CONFIG_NAME_DICTIONARY;
	}

	for (i = 0; i < dlen; i++) {
		struct edict *dict;
		char *index;

		if (prepare_dictionary(dict_names[i]) == -1) {
			return 2;
		}
		index = name_to_index(dict_names[i]);
		dict = edict_open(dict_names[i], index);
		if (dict == NULL) {
			perror(dict_names[i]);
			return 2;
		}
		dicts[i] = dict;
		free(index);
	}

	return do_lookup(dlen, dicts, argc - optind, argv + optind,
			 reverse, reduce, allkana, exact_match,
			 roman, rflags, banner);
}
