/*
 * lookup.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Lookup a word in EDICT.
 *
 */

#include "defs.h"
#include "edict.h"

#include <sys/stat.h>

static char const sccs_id[] = "@(#) myjis lookup.c 1.8 99/01/15 19:18:40";

#define BLOCK_SIZE	512

#define MAX_KEY_SIZE	1024

static int
kana_cmp(wide_char const *w1, wide_char const *w2, size_t len) {
	unsigned long wc1;
	unsigned long wc2 = 0;
	size_t i;

	for(i = 0; i < len; i++) {
		wc1 = (unsigned long) w1[i];
		wc2 = (unsigned long) w2[i];
		
		if (wc1 > wc2) {
			return 1;
		} else if (wc1 != wc2) {
			return -1;
		}
	}
	if (wc2 != '\0'
	    && (w2[i] == H_ya || w2[i] == H_yo || w2[i] == H_yu)) {
		return -1;
	}

	return 0;
}

static wide_char *
get_dictionary_entry(FILE *f, char const *apos) {
	char line_buf[BLOCK_SIZE];
	wide_char *ws;
	long pos = strtol(apos, NULL, 16);

	if (fseek(f, pos, SEEK_SET) == -1) {
		return NULL;
	}
	
	if (fgets(line_buf, BLOCK_SIZE, f) == NULL) {
		return NULL;
	}

	ws = jis_to_wide_dup(line_buf);

	return ws;
}

static void
convert_index_key(wide_char *wd, char const *s, size_t len) {
	static jis_state state;
	

	if (jis_to_wide_str(wd, &s, len, &state) == (size_t) -1) {
		fprintf(stderr,
			"illegal multi-byte sequence in index file.\n");
		exit(1);
	}
	if (s != NULL) {
		fprintf(stderr, "key too long in index file.\n");
		exit(1);
	}
	wide_to_hiragana_str(wd);
	wide_to_hiragana_vowel_str(wd);
}

struct edict *
edict_open(char const *file, char const *index) {
	struct stat st;
	struct edict *ret = xmalloc(sizeof(struct edict));
	int old_errno;

	ret->wkey = NULL;

	ret->dict_file = fopen(file, "rb");
	if (ret->dict_file != NULL) {
		ret->index_file = fopen(index, "rb");
		if (ret->index_file != NULL) {
			if (fstat(fileno(ret->index_file), &st) != -1) {
				ret->blocks = (st.st_size + BLOCK_SIZE - 1)
					      / BLOCK_SIZE;
				return ret;
			}
			old_errno = errno;
			fclose(ret->index_file);
			errno = old_errno;
		}
		old_errno = errno;
		fclose(ret->dict_file);
		errno = old_errno;
	}
	free(ret);
	return NULL;
}

int
edict_close(struct edict *p) {
	if (p->wkey != NULL) {
		free(p->wkey);
	}
	fclose(p->dict_file);
	fclose(p->index_file);
	free(p);
	return 0;
}

	
int
edict_lookup(struct edict *edict, char const *key) {
	FILE *xf = edict->index_file;
	int blocks = edict->blocks;
	int cur_blk;
	int delta;
	char *s;
	wide_char *wkey;
	wide_char wbuf[MAX_KEY_SIZE+1];
	char line_buf[BLOCK_SIZE];
	int c;
	int highest, lowest;
	enum jis_locale_ctypes oldloc;

	wkey = jis_to_wide_dup(key);
 	if (wkey == NULL) {
		return -1;
	}

	wide_to_hiragana_str(wkey);
	wide_to_hiragana_vowel_str(wkey);

	if (edict->wkey != NULL) {
		free(edict->wkey);
	}
	edict->wkey = wkey;

	oldloc = jis_set_locale_ctype(JIS_LOCALE_QUERY);
	jis_set_locale_ctype(JIS_LOCALE_EUC);

	highest = blocks + 1;
	lowest = -1;
	cur_blk = blocks / 2;
	do {
		fseek(xf, cur_blk * BLOCK_SIZE, SEEK_SET);
		if (fgets(line_buf, BLOCK_SIZE, xf) == NULL) {
			break;
		}
		if (fgets(line_buf, BLOCK_SIZE, xf) == NULL) {
			if (!ferror(xf) && feof(xf) && cur_blk == blocks - 1) {
				if (cur_blk != 0) {
					cur_blk--;
				}
				clearerr(xf);
			}
			break;
		}
#if 0
		fprintf(stderr, "@@@ %d \"%.*s\"\n", cur_blk,
			(int) strlen(line_buf) - 1, line_buf);
#endif
		s = strchr(line_buf, ' ');
		if (s == NULL) {
			jis_set_locale_ctype(oldloc);
			errno = EINVAL;
			return -1;
		}
		*s = '\0';

		convert_index_key(wbuf, line_buf, MAX_KEY_SIZE + 1);

		if (wide_cmp(wkey, wbuf) <= 0) {
			delta = (cur_blk - lowest) / 2;
			if (delta == 0 && cur_blk > lowest && cur_blk > 0) {
				if (cur_blk - 1 == lowest) {
					cur_blk--;
					break;
				}
				delta = 1;
			}
			highest = cur_blk;
			cur_blk -= delta;
		} else {
			delta = (highest - cur_blk) / 2;
			if (delta == 0 && cur_blk + 1 < highest) {
				delta = 1;
			}
			lowest = cur_blk;
			cur_blk += delta;
		}
	} while(delta != 0);

	jis_set_locale_ctype(oldloc);

#if 0
	fprintf(stderr, "@@@ highest: %d, lowest %d\n", highest, lowest);
#endif

	if (ferror(xf)) {
		return -1;
	}
	if (feof(xf)) {
		errno = EINVAL;
		return -1;
	}
	
	fseek(xf, cur_blk * BLOCK_SIZE, SEEK_SET);
	if (cur_blk != 0) {
		do {
			c = getc(xf);
			if (c == EOF) {
				if (ferror(xf)) {
					return -1;
				}
				return 0;
			}
		} while(c != '\n');
	}

	return 0;
}

void
edict_rewind(struct edict *edict) {
	free(edict->wkey);
	edict->wkey = NULL;
	rewind(edict->dict_file);
}

char *
edict_get_entry(struct edict *edict, int exact_match, int ending_match) {
	FILE *xf = edict->index_file;
	FILE *df = edict->dict_file;	
	wide_char *wkey = edict->wkey;
	char *s;
	size_t wkey_len;
	int n;
	wide_char ending_char;
	enum jis_locale_ctypes oldloc;
	wide_char *entry;
	wide_char wbuf[MAX_KEY_SIZE+1];
	char line_buf[BLOCK_SIZE];

	oldloc = jis_set_locale_ctype(JIS_LOCALE_QUERY);
	jis_set_locale_ctype(JIS_LOCALE_EUC);

	if (wkey == NULL) {
		char *line;

		line = NULL;
		if (fgets(line_buf, BLOCK_SIZE, df) != NULL) {
			line = strdup(line_buf);
		}
		if (feof(df)) {
			errno = ENOENT;
		}
		jis_set_locale_ctype(oldloc);
		return line;
	}

	wkey_len = wide_len(wkey);
 	if (wkey_len == 0) {
		ending_char = 0;
	} else {
		ending_char = wkey[wkey_len - 1];
	}

#if 0
	fprintf(stderr, "@@@ wkey_len: %ld\n", (long) wkey_len);
#endif

	while(fgets(line_buf, BLOCK_SIZE, xf) != NULL) {
#if 0
		fprintf(stderr, "@@@ \"%.*s\"\n",
			(int) strlen(line_buf) - 1, line_buf);
#endif
		s = strchr(line_buf, ' ');
		if (s == NULL) {
			errno = EINVAL;
			break;
		}
		*s++ = '\0';
		convert_index_key(wbuf, line_buf, MAX_KEY_SIZE + 1);
		if (exact_match) {
			n = wide_cmp(wkey, wbuf);
		} else {
			n = kana_cmp(wkey, wbuf, wkey_len);
		}
		if (n == 0) {
			wide_char *end_of_word;

#if 0
			fputs(s, stdout);
#endif

			entry = get_dictionary_entry(df, s);
			if (entry == NULL) {
				if (feof(df)) {
					errno = EINVAL;
				}
				break;
			}

			end_of_word = wide_chr(entry, (wide_char) ' ');
			if (end_of_word == NULL || end_of_word == entry) {
				errno = EINVAL;
				break;
			}

			if (!ending_match || end_of_word[-1] == ending_char) {
				char *ret;

				jis_set_locale_ctype(oldloc);
				ret = wide_to_jis_dup(entry);
				free(entry);
				return ret;
			}
			free(entry);
		}
		if (n < 0) {
			errno = ENOENT;
			break;
		}
	}

	if (feof(xf)) {
		errno = ENOENT;
	}

	return NULL;
}

