/*
 * myrk.c
 *
 * By Ross Ridge
 * Public Domain
 *
 * Romanji to kana (and vice versa) conversion utility.
 *
 */

#include "defs.h"

#include <unistd.h>

#include "version.c"

static char const myrk_sccs_id[] = "@(#) myjis myrk.c 1.11 99/01/15 19:18:41";

static wide_char *
conv_to_roman(wide_char *ws, int *err, int long_conv, int flags) {
	size_t len;

	if (long_conv) {
		wide_to_hiragana_vowel_str(ws);
	}

	len = 128;
	while(1) {
		size_t tlen;
		wide_char *ret;
		wide_char const *tmp;

		ret = xmalloc(len * sizeof(wide_char));
		tmp = ws;
		tlen = len;
		*err = wide_kana_to_roman_str(ret, &tlen, &tmp, flags);
		if (*tmp == (wide_char) 0) {
			return ret;
		}
		free(ret);
		len *= 2;
	}
}

static int
conv_to_kana(wide_char **list, size_t len, wide_char *ws, int *err,
	     int hiragana, int long_conv, int flags) {
	int i;
	int n;

	n = wide_roman_to_kana_list(list, len, ws, flags, err);
	if (n == -1) {
		return -1;
	}
	for(i = 0; i < n; i++) {
		if (hiragana) {
			wide_to_hiragana_str(list[i]);
			if (long_conv) {
				wide_to_hiragana_vowel_str(list[i]);
			}
		} else if (long_conv) {
			wide_to_katakana_vowel_str(list[i]);
		}
	}
	return n;
}

int
main(int argc, char **argv) {
	int len;
	int ret = 0;
	int i, j;
	char *s;
	int c;
	int flags = 0;
	enum jis_locale_ctypes loc = JIS_LOCALE_DEFAULT;
	int hiragana = 0;
	int long_conv = 1;
	int warn = 0;
	int to_romanji = 0;
	int quiet = 0;
	int all_ambiguous = 0;
		
	setlocale(LC_ALL, "");

	do {
		c = getopt(argc, argv, "ajseuHNKkohlwrqxV");
		switch(c) {
		case 'a':
			all_ambiguous = 1;
			break;

		case 'j':
			loc = JIS_LOCALE_JIS;
			break;

		case 's':
			loc = JIS_LOCALE_SHIFT_JIS;
			break;

		case 'e':
			loc = JIS_LOCALE_EUC;
			break;

#ifdef __JIS_CONFIG_HAVE_UNICODE
		case 'u':
			loc = JIS_LOCALE_UTF8;
			break;
#else /* __JIS_CONFIG_HAVE_UNICODE */
	        case 'u':
			fprintf(stderr, "%s: Sorry, UTF-8 encoding is not"
					" supported.", argv[0]);
			return 1;
#endif /* __JIS_CONFIG_HAVE_UNICODE */
		
		case 'H':
			flags |= JIS_R2K_HEPBURN;
			break;

		case 'N':
			flags |= JIS_R2K_NIPPON;
			break;
	
		case 'K':
			flags |= JIS_R2K_KUNREI;
			break;

		case 'k':
			flags |= JIS_R2K_KATAKANA;
			break;

		case 'o':
			flags |= JIS_R2K_OBSOLETE;
			break;

		case 'x':
			flags |= JIS_R2K_EXTENDED;
			break;

		case 'h':
			hiragana = 1;
			break;

		case 'l':
			long_conv = 0;
			break;

		case 'w':
			warn = 1;
			break;

		case 'r':
			to_romanji = 1;
			break;

		case 'q':
			warn = quiet = 1;
			break;

		case 'V':
			version(myrk_sccs_id);
			return 0;

		case '?':
			fprintf(stderr,
				"usage: %s [-ajseuHNKkohlwrqxV]"
				" [romanji...]\n",
				argv[0]);
			exit(1);
		}
	} while(c != EOF);
			
	jis_set_locale_ctype(loc);

	for (i = optind; i < argc; i++) {
		wide_char *ws;
		wide_char *list[CONFIG_MAX_AMBIGUOUS];
		int err;

		ws = jis_to_wide_dup(argv[i]);
		if (ws == NULL) {
			if (errno == EILSEQ) {
				fprintf(stderr,
					"%s: illegal multibyte sequence"
					" in argument %d.\n", argv[0], i);
				ret = 1;
				continue;
			}
			perror(argv[i]);
			return 1;
		}

		if (to_romanji) {
			list[0] = conv_to_roman(ws, &err, long_conv, flags);
			len = 1;
		} else {
			size_t reqlen = 1;
			if (all_ambiguous) {
				reqlen = CONFIG_MAX_AMBIGUOUS;
			}
			len = conv_to_kana(list, reqlen, ws, &err,
					   hiragana, long_conv, flags);
			if (len == -1) {
				perror("conv_to_kana");
			}
		}

		free(ws);

		for (j = 0; j < len; j++) {
			s = wide_to_jis_dup(list[j]);
			if (s == NULL) {
				if (errno == EILSEQ) {
					fprintf(stderr,
						"%s: Japanese characters not"
						" representable in current"
						" locale.\n", argv[0]);
					return 1;
				}
				perror(argv[i]);
				return 1;
			}
			puts(s);
			free(s);
		}

		if ((err & JIS_R2K_ERR_UNKNOWN) && warn) {
			if (quiet) {
				ret = 1;
			} else {
				fprintf(stderr, "%s: warning: not all"
						" characters were "
						" transliterated.\n", argv[i]);
			}
		}

		if (err & JIS_R2K_ERR_AMBIGUOUS) {
			if (all_ambiguous && !quiet) {
				fprintf(stderr,
					"%s: argument too ambiguous to list"
					" all possibilities.\n",
					argv[i]);
				ret = 1;
			} else if (warn) {
				if (quiet) {
					ret = 1;
				} else {
					fprintf(stderr,
						"%s: warning: ambiguous"
						" romanji in argument.\n",
						argv[i]);
				}
			}
		}
	}
		
	return ret;
}

