/*
 * pts.c
 *
 * By Ross Ridge
 * Public Domain
 *
 */

#include "defs.h"
#include "pts.h"

#if 1
extern "C" {
	int grantpt(int);
	char *ptsname(int);
	int unlockpt(int);
	int getopt(int, char * const *, char const *);
};
#undef NULL
#endif

#include <signal.h>
#include <fcntl.h>
#include <stropts.h>
#include <sys/stream.h>
#include <sys/wait.h>

#if 0
#include <ctype.h>
#include <stdarg.h>
#endif

#ifndef NULL
#define NULL 0
#endif

#include "version.c"

static char const pts_sccs_id[] = "@(#) myjis pts.c 1.7 99/08/25 18:33:06";

#define PTMX_NAME	"/dev/ptmx"

struct termios parent_termios;

void
pt_link::pending::dequeue() {
	struct elt *p = head;

	while(p != NULL) {
		elt *pnext = p->next;
		free(p->data);
		free(p);
		p = pnext;
	}
	head = tail = NULL;
}

ssize_t
pt_link::pending::read_data(char **buf_ptr) {
	if (in_fd == -1) {
		*buf_ptr = NULL;
		return 0;
	}
	static char buf[1024];
	*buf_ptr = buf;
	return read(in_fd, buf, sizeof buf);
}

ssize_t
pt_link::pending::write_data(char *data, ssize_t len) {
	ssize_t ret = write(out_fd, data, len);
	if (ret == -1 && errno == EAGAIN) {
		log("@@@ write: EAGAIN\n");
		ret = 0;
	}
	log("@@@ wrote %ld of %ld on fd %d\n", (long) ret, (long) len, out_fd);
	log("@@@ data = \"");
	log_buf(data, len);
	log("\"\n");
	return ret;
}

int
pt_link::pending::flush_read() {
	assert(0); /* @@@ */
	if (in_fd != -1 && ioctl(in_fd, I_FLUSH, FLUSHR) == -1) {
		lerror("ioctl(I_FLUSH, FLUSHR)");
		return -1;
	}
	return 0;
}	

int
pt_link::pending::flush_write() {
	if (out_fd != -1 && ioctl(out_fd, I_FLUSH, FLUSHW) == -1) {
		lerror("ioctl(I_FLUSH, FLUSHW)");
		return -1;
	}
	return 0;
}	

int
pt_link::pending_out::flush_write() {
	if (out_fd != -1 && tcflush(out_fd, TCOFLUSH) == -1) {
		lerror("tcflush(TCOFLUSH)");
		return -1;
	}
	return 0;
}

int
pt_link::pending_in::flush_read() {
	assert(0); /* @@@ */
	if (in_fd != -1 && tcflush(in_fd, TCIFLUSH) == -1) {
 		lerror("tcflush(TCIFLUSH)");
		return -1;
	}
	return 0;
}

int
pt_link::pending::flush(enum flush_arg arg) {
	dequeue();
	if ((arg == FLUSH_WRITE || arg == FLUSH_BOTH)
	    && flush_write() == -1) {
		return -1;
	}
	if ((arg == FLUSH_READ || arg == FLUSH_BOTH)
	    && flush_read() == -1) {
		return -1;
	}
	return 0;
}
	
int
pt_link::pending::enqueue(char const *s, ssize_t l) {
	if (tail != NULL && l != 0 && tail->len != 0) {
		ssize_t new_len = tail->len + l;
		if (new_len < 256) {
			char *buf = (char *) realloc(tail->data, new_len);
			if (buf == NULL) {
				return -1;
			}
			memcpy(buf + tail->len, s, l);
			tail->data = buf;
			tail->len = new_len;
			return 0;
		}
	}

	struct elt *p = new elt;
	if (p == NULL) {
		return -1;
	}
	if (l == 0) {
		p->data = (char *) malloc(1);
	} else {
		p->data = (char *) malloc(l);
		if (p->data == NULL) {
			delete p;
			return -1;
		}
	}

	memcpy(p->data, s, l);
	p->len = l;
	p->next = NULL;
	if (tail == NULL) {
		head = tail = p;
	} else {
		tail->next = p;
		tail = p;
	}

	return 0;
}

int
pt_link::pending::check_poll() {
	struct pollfd pfd;

	if (!is_pending()) {
		if (in_fd != -1) {
			pfd.events = POLLIN;
			pfd.fd = in_fd;
		} else {
			return 0;
		}
	} else {
		if (out_fd != -1) {
			pfd.events = POLLOUT;
			pfd.fd = out_fd;
		} else {
			return 1;
		}
	}

	int r = poll(&pfd, 1, 0);
	if (r == -1) {
		lerror("poll");
		return -1;
	}
#if 0
	log("@@@ check_poll returns %d on fd %d, event %d\n", 
	    r, pfd.fd, pfd.events);
#endif
	return r;
}

int
pt_link::pending::set_poll() {
	assert(poll_fd == -1);

	int sig_flag;
	if (is_pending()) {
		poll_fd = out_fd;
		sig_flag = S_OUTPUT;
	} else {
		poll_fd = in_fd;
		sig_flag = S_INPUT;
	}

	if (poll_fd != -1) {
		if (ioctl(poll_fd, I_SETSIG, sig_flag) == -1) {
			lerror("ioctl(I_SETSIG)");
			poll_fd = -1;
			return -1;
		}
		log("@@@ I_SETSIG: fd %d, flag %d\n", poll_fd, sig_flag);
	}
	int ret = check_poll();
	if (ret == 1) {
		if (clear_poll() == -1) {
			return -1;
		}
	}
	return ret;

}

int
pt_link::pending::clear_poll() {
	if (poll_fd != -1) {
		if (ioctl(poll_fd, I_SETSIG, 0) == -1 && errno != EINVAL) {
			lerror("ioctl(I_SETSIG, 0)");
			return -1;
		}
		poll_fd = -1;
	}
	return 0;
}

int
pt_link::pending::process() {
	if (!check_poll()) {
		return 0;
	}

	struct elt *p = head;
	if (p != NULL) {
		do {
			ssize_t rlen = p->len;
			ssize_t wlen = write_data(p->data, rlen);
			if (wlen == -1) {
				log("@@@ error writing to fd %d\n", out_fd);
				lerror("write_data");
				return -1;
			}
			if (wlen != rlen) {
				p->len = rlen - wlen;
				memmove(p->data, p->data + wlen, p->len);
				return 1;
			}
			free(p->data);
			p = p->next;
			free(head);
			head = p;
		} while(p != NULL);
		tail = NULL;
		return 1;
	}

	char *buf;
	ssize_t rlen = read_data(&buf);
	if (rlen == -1) {
#if 0
		if (errno == EINTR) {
			return 0;
		}
#endif

		if (errno == EAGAIN) {
			log("@@@ EAGAIN on fd %d\n", in_fd);
			return 0;
		}
		lerror("read_data");
		return -1;
	}
	if (buf == NULL) {
		return 1;
	}

	ssize_t wlen = write_data(buf, rlen);
	if (wlen == -1) {
		log("@@@ error writing to fd %d\n", out_fd);
		lerror("write_data");
		return -1;
	}
	if (wlen != rlen) {
		if (enqueue(buf + wlen, rlen - wlen) == -1) {
			return -1;
		}
	}
	return 1;
}

pt_link::pt_link(int ti, int to, int use_statline)
	: master(-1), slave(-1), ttyin(ti), ttyout(to),
	  pending_output(canon_input),
	  pending_canon_output(),
	  pending_status_line(pending_output),
	  canon_input(&parent_termios,
		      pending_canon_output, pending_input, pending_output,
		      pending_status_line, kdict, use_statline),
	  pending_input(canon_input) {
	return;
}

int
pt_link::open(char const *kdict_name, char const *index_name) {
	if (kdict_name != NULL) {
		if (kdict.open(kdict_name) != 0) {
			lerror(kdict_name);
		} else {
			if (index_name != NULL) {
				if (kdict.read_index(index_name) != 0) {
				lerror(index_name);
				}
			}
			if (kdict.length() < 3) {
				printf("loading %s", kdict_name);
				if (kdict.make_index() != 0) {
					lerror("make_index");
				}
				putchar('\n');
				if (index_name != NULL
				    && kdict.write_index(index_name) != 0) {
					lerror(index_name);
				}
			}
		}
	}

	master = ::open(PTMX_NAME, O_RDWR);
	if (master == -1) {
		lerror(PTMX_NAME);
		return -1;
	}

	if (grantpt(master) == -1) {
		lerror("grantpt");
	} else if (unlockpt(master) == -1) {
		lerror("unlockpt");
	} else {
		char const *slave_name = ptsname(master);
		if (slave_name == NULL) {
			lerror("ptsname");
		} else {
			slave = ::open(slave_name,
				       O_RDWR | O_NOCTTY | O_NONBLOCK);
			if (slave == -1) {
				lerror(slave_name);
			} else {
				if (ioctl(master, I_PUSH, "pckt") == -1) {
					lerror("pckt");
				} else if (ioctl(slave, I_PUSH, "ptem")
					   == -1) {
					lerror("ptem");
				} else if (ioctl(slave, I_PUSH, "ldterm")
					   == -1) {
					lerror("ldterm");
				} else if (ioctl(slave, I_PUSH, "ttcompat")
					   == -1) {
					lerror("ttcompat");
				} else if (ioctl(master, TIOCREMOTE, 1)
					   == -1) {
					lerror("TIOCREMOTE");
				} else {
					return 0;
				}
				::close(slave);
				slave = -1;
			}
		}
	}

	::close(master);
	master = -1;
	return -1;
}

int
pt_link::open_slave() {
	assert(slave != -1);
	::close(slave);
	slave = -1;
	char const *slave_name = ptsname(master);
	if (slave_name == NULL) {
		lerror("ptsname");
	} else {
		int child_slave = ::open(slave_name, O_RDWR);
		if (child_slave == -1) {
			lerror(slave_name);
		} else {
			return child_slave;
		}
	}
	return -1;
}

ssize_t
pt_link::pending_in::read_data(char **buf_ptr) {
	char *buf;
	ssize_t len = pending::read_data(&buf);

	if (!canon_input.raw()) {
		int i;
		for(i = 0; i < len; i++) {
			int ret;
			ret = canon_input.process_canon_input(buf[i]);
			if (ret != 0) {
				return ret;
			}
		}
		*buf_ptr = NULL;
		return 0;
	}
	*buf_ptr = buf;
	return len;
}

int
pt_link::pending_out::process_ioctl(int cmd, char *data, size_t len) {
	switch(cmd) {
	case TCSETA:
 	case TCSETAW:
	case TCSETAF:
		struct termio otio;
		assert(len >= sizeof otio);
		memcpy(&otio, data, sizeof otio);
		return canon_input.process_termio(cmd, &otio);

	case TCSBRK:
	case TIOCGWINSZ:
	case TIOCSWINSZ:
		log("ioctl 0x%X ignored\n", cmd);
		return 0;

	case TCSETS:
	case TCSETSW:
	case TCSETSF:
		struct termios tio;
		assert(len >= sizeof tio);
		memcpy(&tio, data, sizeof tio);
		log("tio.c_cc[0] = %d, tio.c_cc[1] = %d, "
		    "tio.c_cc[2] = %d, tio.c_cc[3] = %d\n",
		    (int) tio.c_cc[0], (int) tio.c_cc[1],
		    (int) tio.c_cc[2], (int) tio.c_cc[3]);
		return canon_input.process_setattr(cmd, &tio);

	case TCGETS:
	case TCGETA:
	case TCXONC:
	case TCFLSH:
	default:
		log("unexpected ioctl: 0x%X\n", cmd);
		return -1;
	}
}

ssize_t
pt_link::pending_out::read_data(char **buf_ptr) {
	static char buf[2048];
	unsigned char msg;
	struct strbuf ctl, data;

	ctl.maxlen = 1;
	ctl.buf = (char *) &msg;
	data.maxlen = sizeof buf;
	data.buf = buf;
	int flags = 0;

	ssize_t r = getmsg(in_fd, &ctl, &data, &flags);
	if (r == -1) {
		lerror("getmsg");
		return -1;
	}

	if (msg != 0) {
		log("ctl.buf[0] = 0x%02X, data.len = %d\n",
		    (unsigned char) msg, data.len);
	}

	if (r != 0) {
		log("incomplete message (%d).\n", r);
		if (r & MORECTL) {
			return -1;
		}
	}

	ssize_t len = data.len;

	switch(msg) {
	case M_FLUSH:
		if (flush(FLUSH_WRITE) == -1) {
			return -1;
		}
		break;

	case M_IOCTL:
		struct iocblk ioc;
		assert((size_t) len >= sizeof ioc);
		memcpy(&ioc, buf, sizeof ioc);
		log("ioc_cmd: %x, ioc_id %u, ioc_count %u,"
		    " ioc_error %d, ioc_rval = %d\n",
		    (unsigned) ioc.ioc_cmd, (unsigned) ioc.ioc_id,
		    (unsigned) ioc.ioc_count, (int) ioc.ioc_error,
		    (int) ioc.ioc_rval);
		assert((size_t) len >= sizeof ioc + ioc.ioc_count);
		if (process_ioctl(ioc.ioc_cmd, buf + sizeof ioc,
				  ioc.ioc_count)) {
			return -1;
		}
		break;

	case M_DATA:
		*buf_ptr = buf;
		return len;
	}

	*buf_ptr = NULL;
	return 0;
}

ssize_t
pt_link::pending_out::write_data(char *data, ssize_t len) {
	ssize_t ret = pending::write_data(data, len);
	ssize_t i;

	if (ret == -1) {
		state = NORMAL;
		return ret;
	}

	/*
	 * Keep track of ANSI X3.64 style escape sequences
	 */
	for(i = 0; i < ret; i++) {
		char c = data[i];
		switch(state) {
		case NORMAL:
			if (c == '\033') {
				state = ESC;
			} else if (c == '\233') {
				state = PARAM;
			}
			break;
		case ESC:
			if (c == '[') {
				state = PARAM;
			} else {
				state = NORMAL;
			}
			break;
		case PARAM:
			c = c & 0xf0;
			if (c == 0x20) {
				state = FINAL;
			} else if (c != 0x30) {
				state = NORMAL;
			}
			break;
		case FINAL:
			if (c & 0xf0 != 0x20) {
				state = NORMAL;
			}
			break;
		}
	}
	return ret;
}

ssize_t
pt_link::pending_statline::write_data(char *data, ssize_t len) {
	if (pending_output.status_line_ok()) {
		if (pending_output.enqueue(data, len) == -1) {
			return -1;
		}
		return len;
	}
	return 0;
}

int
pt_link::close() {
	::close(master);
	::close(slave);
	return 0;
}

extern "C" void
null_signal_handler(...) {
	return;
}

#if 0
inline int
test_poll(struct pollfd const &pans) {
	return pans.revents & pans.events;
}
#endif

int
pt_link::process(pid_t child, int *status) {
	sigset_t old_sigmask;

	if (sigprocmask(SIG_SETMASK, NULL, &old_sigmask) == -1) {
		lerror("sigprocmask(SIG_SETMASK, NULL)");
		return 1;
	}

	sigset_t suspend_sigmask = old_sigmask;
	sigdelset(&suspend_sigmask, SIGCHLD);
	sigdelset(&suspend_sigmask, SIGPOLL);

	sigset_t new_sigmask = old_sigmask;
	sigaddset(&new_sigmask, SIGCHLD);
	sigaddset(&new_sigmask, SIGPOLL);

	if (sigprocmask(SIG_SETMASK, &new_sigmask, NULL) == -1) {
		lerror("sigprocmask(SIG_SETMASK, &new)");
		return 1;
	}

	struct sigaction signal_action;
	signal_action.sa_handler = null_signal_handler;
	signal_action.sa_mask = new_sigmask;
	signal_action.sa_flags = 0;

	struct sigaction old_sigchld_action;
	if (sigaction(SIGCHLD, &signal_action, &old_sigchld_action) == -1) {
		lerror("sigaction(SIGCHLD)");  
		return 1;
	}

	struct sigaction old_sigpoll_action;
	if (sigaction(SIGPOLL, &signal_action, &old_sigpoll_action) == -1) {
		lerror("sigaction(SIGPOLL)");
		return 1;
	}

	/*
	 * Check to see if the child process has already died.
	 */
	int r = waitpid(child, status, WNOHANG);
	if (r != 0) {
		if (r == -1) {
			lerror("waitpid");
			return 1;
		}
		return 0;
	} 

	pending_output.set_fds(master, ttyout);
	pending_input.set_fds(ttyin, master);
	pending_canon_output.set_fds(-1, slave);
	canon_input.init();

	int ttyin_flags = fcntl(ttyin, F_GETFL);
	if (ttyin_flags == -1) {
		lerror("F_GETFL");
		return 1;
	}
	if (fcntl(ttyin, F_SETFL, ttyin_flags | O_NONBLOCK) == -1) {
		lerror("F_SETFL");
		return 1;
	}

	int ret;
	int shutdown = 0;

	while(1) {
		log("@@@ checking for I/O\n");

		int do_pause;

		ret = pending_output.set_poll();
		if (ret == -1) {
			break;
		}
		do_pause = !ret;

		if (do_pause) {
			ret = pending_canon_output.set_poll();
			if (ret == -1) {
				break;
			}
			do_pause = !ret;
		}

		if (!shutdown && do_pause) {
			ret = pending_input.set_poll();
			if (ret == -1) {
				break;
			}
			do_pause = !ret;
		}

		if (do_pause && (!shutdown || pending_output.is_pending())) {
			ret = sigsuspend(&suspend_sigmask);
			if (ret == -1 && errno != EINTR) {
				lerror("sigsuspend");
				break;
			}
		}

		ret = pending_output.clear_poll();
		if (ret == -1) {
			break;
		}
		ret = pending_input.clear_poll();
		if (ret == -1) {
			break;
		}
		ret = pending_canon_output.clear_poll();
		if (ret == -1) {
			break;
		}

		sigprocmask(SIG_SETMASK, &suspend_sigmask, NULL);
		sigprocmask(SIG_SETMASK, &new_sigmask, NULL);

		log("@@@ processing\n");

		if (!shutdown) {
			ret = waitpid(child, status, WNOHANG);
			if (ret != 0) {
				log("waitpid returned %d\n", ret);
				if (ret == -1) {
					lerror("waitpid");
					break;
				}
				log("status = 0x%04x\n", *status);

				shutdown = 1;
#if 0
				::close(slave);
				slave = -1;
#endif
			}
		}

		if (!shutdown) {
			ret = pending_input.process();
			if (ret == -1) {
				break;
			}
		}

		ret = pending_canon_output.process();
		if (ret == -1) {
			break;
		}

		ret = pending_status_line.process();
		if (ret == -1) {
			break;
		}

		ret = pending_output.process();
		if (ret == -1) {
			break;
		}
		if (shutdown && ret == 0) {
			break;
		}
	} 

	pending_output.clear_poll();
	pending_input.clear_poll();
	pending_canon_output.clear_poll();

	if (fcntl(ttyin, F_SETFL, ttyin_flags) == -1) {
		lerror("F_SETFL");
		return 1;
	}

	r = sigaction(SIGCHLD, &old_sigchld_action, NULL);
	if (r == -1) {
		lerror("sigaction(old_chld)");
		return 1;
	}

	r = sigaction(SIGCHLD, &old_sigpoll_action, NULL);
	if (r == -1) {
		lerror("sigaction(old_poll)");
		return 1;
	}

	if (sigprocmask(SIG_SETMASK, &old_sigmask, NULL) == -1) {
		lerror("sigprocmask(SIG_SETMASK, &old)");
		return 1;
	}

	return ret;
}

pt_link::~pt_link() {
	close();
}

int
main(int argc, char **argv) {
	setlocale(LC_ALL, "");
	jis_set_locale_ctype(JIS_LOCALE_EUC);

	char *index = NULL;
	char *kdict = NULL;
	int c;
	do {
		c = getopt(argc, argv, "l:k:i:");
		switch(c) {
		case 'i':
			index = optarg;
			break;

		case 'k':
			kdict = optarg;
			break;

		case 'l':
			flog = fopen(optarg, "w");
			if (flog == NULL) {
				perror(optarg);
				return 1;
			}
			setvbuf(flog, NULL, _IOLBF, 0);
			break;

		case '?':
			fprintf(stderr, 
"usage: %s [-k kdict [-i kdict.index]] [-l logfile]\n",
				argv[0]);
			return 1;
		}
	} while(c != EOF);
	
	version(pts_sccs_id);

	if (ioctl(0, TCGETS, &parent_termios) == -1) {
		lerror("stdin");
	}

	pt_link pt(0, 1);

	if (pt.open(kdict, index) != 0) {
		return 1;
	}

	fflush(stdout);
	fflush(stderr);

	pid_t slave_pid = fork();
	switch(slave_pid) {
	case -1:
		lerror("fork");
		return 1;
		
	case 0:
		if (setsid() == -1) {
			lerror("setsid");
			return 1;
		}
		if (setpgrp() == -1) {
			lerror("setpgrp");
			return 1;
		}

		int slave = pt.open_slave();
		if (slave == -1) {
			return 1;
		}

		struct termios tio = parent_termios;
		tio.c_cc[VSWTCH] = '\001';
		if (tcsetattr(slave, TCSANOW, &tio) == -1) {
			lerror("ioctl child");
			return 1;
		}

		int old_stderr = dup(2);
		if (old_stderr != -1) {
			if (fcntl(old_stderr, F_SETFD, 1) == -1) {
				lerror("fcntl");
				return 1;
			}
		}  

		close(0);
		close(1);
		close(2);
		dup(slave);
		dup(slave);
		dup(slave);
		close(slave);
		pt.close();

		char const *cmd;
		if (optind != argc) {
			cmd = argv[optind];
			execvp(cmd, argv + optind);
		} else {
			cmd = "/bin/sh";
			execl(cmd, cmd, NULL);
		}

		if (old_stderr != -1) {
			close(2);
			dup(old_stderr);
			lerror(cmd);
			fflush(stderr);
		}	
		_exit(1);
	}

	int wait_st;
	int ret = pt.process(slave_pid, &wait_st);

	if (ioctl(0, TCSETS, &parent_termios) == -1) {
		lerror("stdin");
	}

	return ret;
}

/* Local variables: */
/* mode: c++ */
/* End: */
