/*
 * pts_ci.c
 *
 * By Ross Ridge
 * Public Domain
 *
 */

#include "defs.h"
#include "pts.h"

#include <ctype.h>
#include <signal.h>

static char const sccs_id[] = "@(#) myjis pts_ci.c 1.2 00/05/15 14:03:08";

#define MESSAGE_OFFSET	15
#define SCREEN_COLUMNS	80
#define MESSAGE_LENGTH	(SCREEN_COLUMNS - MESSAGE_OFFSET - 1)
#define MESSAGE_SCROLL	(MESSAGE_LENGTH - 4)

const int pt_link::canon_in::euc_byte_len[4] = {
	1, 2, 2, 3
};

const int pt_link::canon_in::euc_char_width[4] = {
	1, 2, 1, 2
};

pt_link::canon_in::canon_in(struct termios *parent_termios,
			    pending &co, pending_in_base &in, pending_out &out,
			    pending_statline &sl,
			    kanjidict &kd, int use_sl, int mode, int recurse)
	: canon_pos(0), backslash(0), lnext(0),
	  command_mode(NONE), input_mode((enum imode) mode),
	  ignore_bytes(0), use_status_line(use_sl), recursive_edit(recurse),
	  status_line_prompt(NULL), sl_msg_pos(0), sl_msg_offset(0),
	  translation_pos(0),
          r2k_flags(JIS_R2K_ALL | JIS_R2K_KATAKANA | JIS_R2K_PARTICLE
		    | JIS_R2K_OBSOLETE | JIS_R2K_EXTENDED),
	  r2k_kana_count(0), r2k_last_substitute(0),
	  master_termios(*parent_termios),
	  pending_canon_output(co), pending_input(in), pending_output(out),
	  pending_status_line(sl),
	  kdict(kd), kinput(NULL) {
	master_termios.c_oflag = 0;
#if 0
	master_termios.c_lflag = ISIG;
#else
	master_termios.c_lflag = 0;
#endif
	master_termios.c_cc[VMIN] = 1;
	master_termios.c_cc[VTIME] = 0;
	return;
}

pt_link::canon_in::~canon_in() { 
	if (kinput != NULL) {
		delete kinput;
		return;
	}
	command_mode = NONE;
	input_mode = NORMAL;
	status_line_prompt = NULL;
	sl_msg_offset = sl_msg_pos = 0;
	update_status_line(1); 
	pending_status_line.process();
	pending_output.process();
}

ssize_t
pt_link::canon_in::pending_fake_in::write_data(char *data, ssize_t len) {
	assert(input == NULL);
	if (input != NULL) {
		errno = EIO;
		return -1;
	}
	input = (char *) malloc(len + 1);
	if (input == NULL) {
		return -1;
	}
	memcpy(input, data, len);
	input[len] = '\0';
	return len;
}

int
pt_link::canon_in::init() {
	if (update_status_line(1) == -1 
	    || pending_status_line.process() == -1
	    || pending_output.process() == -1) {
		return -1;
	}
	return 0;
}

int
pt_link::canon_in::flush_output() {
	/* ??? */
	return 0;
}

int
pt_link::canon_in::status_line_echo(char c) {
	if (c == '\n') {
		return 0;
	} else if (c == '\b') {
		if (sl_msg_pos > 0) {
			sl_msg_pos--;
		}
		return 0;
	}
	if (!euc_is_print(c)) {
		return 0;
	}
	if (sl_msg_pos < (int) sizeof sl_msg_buf) {
		sl_msg_buf[sl_msg_pos++] = c;
	}
	return 0;
}

int
pt_link::canon_in::status_line_echo(char const *s) {
	while(*s != '\0') {
		if (status_line_echo(*s++) != 0) {
			return -1;
		}
	}
	return 0;
}

int
pt_link::canon_in::write_echo(char c) {
	if (recursive_edit && use_status_line) {
		if (status_line_echo(c) == -1 || update_status_line() == -1) {
			return -1;
		}
		return 0;
	}
	int r = pending_canon_output.enqueue(&c, 1);
	if (r != 0) {
		return r;
	}
	if (pending_canon_output.process() == -1) {
		return -1;
	}
	return 0;
}

int
pt_link::canon_in::write_echo(char const *s) {
	if (recursive_edit && use_status_line) {
		if (status_line_echo(s) == -1 || update_status_line() == -1) {
			return -1;
		}
		return 0;
	}
	int r = pending_canon_output.enqueue(s, strlen(s));
	if (r != 0) {
		return r;
	}
	if (pending_canon_output.process() == -1) {
		return -1;
	}
	return 0;
}

int
pt_link::canon_in::update_status_line(int force) {
	static char const to_sl[] = "\033[?E\E[?1T";
	static char const from_sl[] = "\033[?F";
	static char const command_help[] = "[ahkn]";
	char const prompt_offset = 10;

	if (!use_status_line) {
		return 0;
	}

	char line[500];
	static char old_line[500];
	int pos = 0;

	switch(input_mode) {
	case HIRAGANA:
		line[pos++] = '\244';
		line[pos++] = '\253';
		break;

	case KATAKANA:
		line[pos++] = '\245';
		line[pos++] = '\253';
		break;

	default:
		line[pos++] = ' ';
		line[pos++] = ' ';
		break;
	}

	line[pos++] = ' ';

	if (command_mode == COMMAND) {
		memcpy(line + pos, command_help, sizeof command_help - 1);
		pos += sizeof command_help - 1;
	} else if (input_mode != NORMAL
		   && canon_pos > translation_pos
		   && euc_is_print(buf[translation_pos])) {
		int len = canon_pos - translation_pos;
		memcpy(line + pos, buf + translation_pos, len);
		pos += len;
	}

	if (status_line_prompt != NULL || sl_msg_pos > 0) {
		while(pos < prompt_offset) {
			line[pos++] = ' ';
		}
	}

	if (status_line_prompt) {
		strcpy(line + pos, status_line_prompt);
		pos += strlen(status_line_prompt);
	}

	
	size_t msglen = sl_msg_pos - sl_msg_offset;
	if (msglen > 0) {
		if (msglen >= sizeof line - pos) {
			msglen = sizeof line - pos - 1;
		}
		memcpy(line + pos, sl_msg_buf + sl_msg_offset, msglen);
		pos += msglen;
	}

 	log("@@@ sl: len %d, line \"", pos);
	log_buf(line, pos);
	log("\"\n");

	if (pos > 78) {
		pos = 78;
	}

	line[pos++] = '\0';

	if (force || memcmp(line, old_line, pos) != 0) {
		memcpy(old_line, line, pos);
		if (pending_status_line.enqueue(to_sl, sizeof to_sl - 1) == -1
		    || pending_status_line.enqueue(line, pos - 1) == -1
		    || pending_status_line.enqueue(from_sl, sizeof from_sl - 1)
 		       == -1) {
			return -1;
		}
	}

	return 0;
}


int
pt_link::canon_in::print_mode(char const *s) {
	if (use_status_line) {
		return update_status_line();
	}
	if (echo) {
		return write_echo(s);
	}
	return 0;
}

inline int
pt_link::canon_in::flush_prompt() {
	if (use_status_line 
	    && (update_status_line() == -1
	        || pending_status_line.process() == -1)) {
		return -1;
	}
	if (pending_output.process() == -1) {
		return -1;
	}
	return 0;
}

int
pt_link::canon_in::print_prompt(char c) {
	int r;

	if (use_status_line) {
		r = status_line_echo(c);
	} else {
		r = pending_output.enqueue(&c, 1);
	}
	if ((c != '\n' && c != '\a') || r != 0) {
		return r;
	}
	return flush_prompt();
}

int
pt_link::canon_in::print_prompt(char const *s) {
	size_t len = strlen(s);
	if (len == 0) {
		return 0;
	}
	int r;
	if (use_status_line) {
		r = status_line_echo(s);
	} else {
		r = pending_output.enqueue(s, len);
	}
	if (r != 0 || s[len - 1] != '\n') {
		return r;
	}
	return flush_prompt();
}

int
pt_link::canon_in::set_normal_mode() {
	log("@@@ set_normal_mode\n");
	kanji_selection = NULL;
	command_mode = NONE;
	status_line_prompt = NULL;
	sl_msg_offset = sl_msg_pos = 0;
#if 0
	int r = update_status_line();
	if (r != 0) {
		return r;
	}
#endif
	return process_setattr(TCSETS, &old_slave_termios);
}

int
pt_link::canon_in::set_prompt_mode() {
	log("@@@ set_prompt_mode\n");

	if (command_mode == NONE) {
		old_slave_termios = slave_termios;
	}

	command_mode = KANJI;

	if (process_setattr(TCSETS, &parent_termios) == -1) {
		return -1;
	}

	if (use_status_line) {
		status_line_prompt = "KC: ";
		int r = update_status_line();
		if (r != 0) {
			return r;
		}
	} else {
		if (print_prompt("Kanji Command: ") != 0
		    || flush_prompt() != 0) {
			return -1;
		}
	}
	return 0;
}

int
pt_link::canon_in::set_long_prompt_mode() {
	log("@@@ set_long_prompt_mode\n");

	if (command_mode == NONE) {
		old_slave_termios = slave_termios;
	}

	if (kinput == NULL) {
		kinput = new canon_in(&master_termios,
				      pending_output, fake_input,
				      pending_output, pending_status_line,
				      kdict, use_status_line, input_mode, 1);
		if (kinput == NULL) {
			return -1;
		}
	}

	kinput->input_mode = input_mode;
	kinput->status_line_prompt = status_line_prompt;
	kinput->sl_msg_offset = kinput->sl_msg_pos = 0;

	if (kinput->process_setattr(TCSETS, &parent_termios) == -1) {
		return -1;
	}

	command_mode = KANJI_INPUT;
	return 0;
}

int
pt_link::canon_in::process_echo(char c) {
	if (echo) {
		if (slave_termios.c_lflag & ECHOCTL) {
			/*
			 *  Assume ASCII.
			 */
			static char buf[] = "^@";
			if (c >= 0 && c < 32 && c != '\n') {
				buf[1] = c + '@';
				return write_echo(buf);
			} else if (c == 127) {
				buf[1] = '?';
				return write_echo(buf);
			}
		}
		return write_echo(c);
	}
	return 0;
}
				
int
pt_link::canon_in::process_interrupt(char c, int intr) {
	if ((slave_termios.c_lflag & NOFLSH) == 0) {
		/*
		 *  Flush input.
		 */
		translation_pos = canon_pos = 0;
		int r = pending_input.flush(FLUSH_WRITE);
		if (r != 0) {
			return r;
		}
	}

	int r = process_echo(c);
	if (r != 0) {
		return r;
	}

	log("sending interrupt %d\n", intr);

	return pending_input.send_sig(intr);
}

int
pt_link::canon_in::process_reprint() {
	slave_termios.c_lflag &= ~PENDIN;

	int r = process_echo('\n');
	if (r != 0) {
		return r;
	}
	int i;
	for (i = 0; i < canon_pos; i++) {
		r = process_echo(buf[i]);
		if (r != 0) {
			return r;
		}
	}
	return 0;
}

int
pt_link::canon_in::process_erase(int echoe, int dest) {
	int ret = 0;
	int update_status = 0;

#if 0
	log("process_erase at %d, to %d\n", canon_pos, dest);
#endif

	while(canon_pos > dest) {
		if (canon_pos == 0) {
			break;
		}

		/*
		 * Find the start of the previous character.
		 */

		int incomplete_mb_char = 0;
		int char_set = char_marker[--canon_pos];
		if (char_set != -1 && char_set != 1) {
			incomplete_mb_char = 1;
		}
		while(char_set > 0) {
			canon_pos--;
			assert(canon_pos >= 0);
			char_set = char_marker[canon_pos];
		}

		if (canon_pos < translation_pos) {
			translation_pos = canon_pos;
		} else {
			update_status = 1;
		}

		assert(char_set >= -4 && char_set <= -1);

		if (echo) {
			if (echoe) {
				/*
				 * Figure out the display width of the
				 * previous character.
				 */
				if (incomplete_mb_char) {
					/*
					 * Incomplete multibyte character.
					 */
					log("incomplete mb char\n");
					ret = write_echo('\a');
					break;
				}

				int char_width = euc_char_width[-char_set - 1];
				if (char_set == -1
				    && (slave_termios.c_lflag & ECHOCTL)) {
					char c = buf[canon_pos];
					if ((c >= 0 && c < 32) || c == 127) {
						char_width = 2;
					}
				}

				while(char_width-- > 0) {
					ret = write_echo("\b \b");
					if (ret != 0) {
						break;
					}
				}

			} else {
				ret = process_echo(slave_termios.c_cc[VERASE]);
				if (ret != 0) {
					break;
				}
			}
		}
	}
	if (ret == 0 && update_status) {
		return update_status_line();
	}
	return ret;
}

int
pt_link::canon_in::process_break(int pos) {
	log("@@@: input break\n");
	int r = pending_input.enqueue(buf, pos);
	canon_pos -= pos;
	memmove(buf, buf + pos, canon_pos);
	translation_pos -= pos;
	if (translation_pos < 0) {
		translation_pos = 0;
	}
	if (r != 0) {
		return r;
	}
	if (pending_input.process() == -1) {
		return -1;
	}
	return 0;
}

int
pt_link::canon_in::substitute_input(int dest, wide_char const *ws,
				    int advance_pos) {
	
	int r = process_erase(1, dest);
	if (r != 0) {
		return r;
	}

	jis_state tmp = __jis_initial_state;
	wide_char const *wps = ws;
	size_t len = wide_to_jis_str(buf + canon_pos, &wps, 
				     sizeof buf - canon_pos - 1, &tmp);
	if (len == (size_t) -1) {
		lerror("wide_to_jis_str");
		return -1;
	}

	assert(jis_is_initial_state(&tmp));
	if (wps != NULL) {
		log("@@@ kana won't fit\n");
		return write_echo('\a');
	}

	log("@@@: ws = { %08lX, %08lX, %08lX, %08lX } -> \"",
	    (long) ws[0], (long) ws[1], (long) ws[2], (long) ws[3]);
	log_buf(buf + dest, len);
	log("\"\n");
		
	size_t i;
	for(i = 0; i < len; i++) {
		int r = process_character(buf[dest + i], 0);
		if (r != 0) {
			return r;
		}
	}

	if (advance_pos) {
		translation_pos = canon_pos;
	}
	if (icanon || translation_pos == 0) {
		return 0;
	}
	return process_break(translation_pos);
}
	
int
pt_link::canon_in::roman_to_kana(char c) {
	int ret = 1;

	if (c == '\033' && !icanon) {
		input_mode = NORMAL;
		log("@@@ got escape\n");
		if (write_echo('\a') != 0) {
			return -1;
		}
		if (update_status_line() != 0) {
			return -1;
		}
		return 1;
	}
	buf[canon_pos] = c;
	buf[canon_pos + 1] = '\0';

	int possible_completion = 0;
	int pos = translation_pos;
	assert(pos <= canon_pos);
	while(pos <= canon_pos) {
		log("@@@ translating at %d\n", pos);
		jis_state tmp = __jis_initial_state;
		wide_char wbuf[sizeof buf];
		char const *ps = buf + pos;
		size_t len = jis_to_wide_str(wbuf, &ps, sizeof buf, &tmp);
		if (len != (size_t) -1 && len != 0
		    && jis_is_initial_state(&tmp)) {
			/*
			 *  Handle n ambiguity.
			 */
			if (wide_to_katakana(wbuf[0]) == K_N) {
				wbuf[0] = 'n';
			}
			wbuf[len] = (wide_char) 0;
			int count = wide_roman_to_kana(r2k_kana, R2K_KANA_LEN,
						       wbuf, r2k_flags);
			possible_completion = 0;
			if (count == -1) {
				count = 0;
			} else if (count == 0) {
				possible_completion = 1;
			}
			r2k_kana_count = count;
			
			int i;
#if 1
			for (i = 0; i < count; i++) {
				log("@@@: roman '%s'\n", r2k_kana[i]->roman);
			}
#endif
			for (i = 0; i < count; i++) {
				char const *roman = r2k_kana[i]->roman;
				size_t rlen = strlen(roman);
				if (rlen == len) {
					break;
				}
			}
			if (i != count) {
				wide_char kana[__JIS_R2K_MAX_KANA_LEN + 1];
				memcpy(kana, r2k_kana[i]->kana,
				       (__JIS_R2K_MAX_KANA_LEN + 1)
				       * sizeof (wide_char));
				if (input_mode == HIRAGANA) {
					wide_to_hiragana_str(kana);
				}
				/*
				 *  Handle n ambiguity.
				 */
				if (strcmp(r2k_kana[i]->roman, "n") != 0) {
					return substitute_input(pos, kana, 1);
				}
				ret = substitute_input(pos, kana, 0);
				break;
			}
		}
		do {
			pos++;
		} while(pos <= canon_pos && char_marker[pos] >= 0);

		if (!possible_completion) {
			translation_pos = pos;
			log("@@@ no possible completion\n");
		}

#if 0
		if (canon_pos + 1 - pos > __JIS_R2K_MAX_ROMAN_LEN) {
			translation_pos = pos;
		}
#endif
	}

	return ret;
}

int
pt_link::canon_in::process_character(char c, int translate) {
	/*
	 * Figure out what byte of a possibly multi-byte character
	 * this byte is.
	 */
	char marker;
	if (euc_is_ss2(c)) {
		marker = -3;
	} else if (euc_is_ss3(c)) {
		marker = -4;
	} else if (euc_is_mb(c)) {
		int last_marker = 0;
		if (canon_pos > 0) {
			last_marker = char_marker[canon_pos - 1];
			if (last_marker < 0) {
				last_marker = euc_byte_len[-last_marker - 1];
			}
		}
		if (last_marker > 1) {
			marker = last_marker - 1;
		} else {
			marker = -2;
		}
	} else {
		marker = -1;
	}

	if (marker < 0 && canon_pos + euc_byte_len[-marker - 1]
	                  >= (int) sizeof buf - 1) {
		ignore_bytes = euc_byte_len[-marker - 1] - 1;
		if (slave_termios.c_lflag & IMAXBEL) {
			return write_echo('\a');
		}
		return 0;
	}

	char_marker[canon_pos] = marker;

	if (translate && input_mode != NORMAL && marker == -1) {
		int r = roman_to_kana(c);
		if (r == -1) {
			return -1;
		}
		if (r == 0) {
			return update_status_line();
		}
		buf[canon_pos++] = c;
		r = update_status_line();
		if (r != 0) {
			return r;
		}
	} else {
		buf[canon_pos++] = c;
	}

			
	int r = process_echo(c);
	if (r != 0) {
		return r;
	}
	if (icanon || (marker != 1 && marker != -1)) {
		return 0;
	}
	if (input_mode != NORMAL) {
		if (translation_pos != 0) {
			return process_break(translation_pos);
		}
		return 0;
	}
	return process_break();
}

int
pt_link::canon_in::process_command(char c) {
	command_mode = NONE;

	switch(c) {
	case 'n':
		input_mode = NORMAL;
		if (print_mode("[roman on]") != 0) {
			return -1;
		}
		break;

	case 'h':
		input_mode = HIRAGANA;
		translation_pos = canon_pos;
		if (print_mode("[hiragana on]")) {
			return -1;
		}
		break;

	case 'k':
		input_mode = KATAKANA;
		translation_pos = canon_pos;
		if (print_mode("[katakana on]")) {
			return -1;
		}
		break;

	case 'a':
		if (kdict.length() < 3) {
			if (print_mode("[kanji input unavailable]") != 0) {
				return -1;
			}
			return write_echo('\a');
		}
		kanji_selection = NULL;
		if (set_prompt_mode() == -1) {
			return -1;
		}
		break;

	default:
		int r = update_status_line();
		if (r != 0) {
			return r;
		}
		log("@@@ unknown command\n");
		return write_echo('\a');
	}

	return 0;
}

int
pt_link::canon_in::start_kanji_input(char const *prompt, char cmd) {
	if (use_status_line) {
		status_line_prompt = prompt;
		sl_msg_offset = sl_msg_pos = 0;
		if (update_status_line() != 0) {
			return -1;
		}
	} else {
		if (print_prompt(prompt) != 0 || flush_prompt() != 0) {
			return -1;
		}
	}

	kanji_command = cmd;

	return set_long_prompt_mode();
}

int
pt_link::canon_in::kanji_display_selection(int limit) {
	int len = kanji_selection.length();
	assert(len < 100000);
	
	if (use_status_line) {
		limit = len;
	}

	if (len == 0) {
		if (print_prompt("Selection Empty\n") == -1) {
			return -1;
		}
	} else {
		char sbuf[16];
		sprintf(sbuf, "%d:", len);

		if (print_prompt(sbuf) != 0) {
			return -1;
		}

		jis_state state = __jis_initial_state;

		int i;
		for(i = 0; i < len && i < limit; i++) {
			char jbuf[JIS_MB_LEN_MAX + 1];

#if 0
			log("@@@ line %d\n", kanji_selection[i]);
#endif
			wide_char wc = kdict[kanji_selection[i]];
			size_t n = wide_to_jis(jbuf, sizeof jbuf - 1, wc, &state);
			if (n == (size_t) -1) {
				return -1;
			}
			jbuf[n] = '\0';
			assert(jis_is_initial_state(&state));

			if (i < 10) {
				sprintf(sbuf, " %d)", i);
			} else {
				sprintf(sbuf, " %d:", i);
			}
			if (print_prompt(sbuf) != 0
			    || print_prompt(jbuf) != 0) {
				return -1;
			}
		}
		if (print_prompt('\n') != 0) {
			return -1;
		}
	}
	return 0;
}

int
pt_link::canon_in::merge_selection(list<int> const &select) {
	if (select.length() == 0) {
		if (print_prompt("[no match] ") != 0) {
			return -1;
		}
	} else if (kanji_selection.length() == 0) {
		kanji_selection = select;
	} else {
		list<int> tmp_select = select;

		tmp_select.qsort(qvcompare<int>::fn);
		kanji_selection.qsort(qvcompare<int>::fn);

#if 0
		log("tmp_select = [ ");
		log_list_int(tmp_select);
		log(" ]\nkanji_select = [ ");
		log_list_int(kanji_selection);
		log(" ]\n");
#endif

		int ilen = tmp_select.length();
		int jlen = kanji_selection.length();

		list<int> new_select;
		int j = 0;
		int i;
		for(i = 0; i < ilen; i++) {
			int n = tmp_select[i];
			while(j < jlen && kanji_selection[j] < n) {
				j++;
			}
			if (j == jlen) {
				break;
			}
			if (n == kanji_selection[j]) {
				if (new_select.add(n) != 0) {
					return -1;
				}
			}
		}
		if (new_select.length() == 0) {
			kanji_selection = select;
		} else {
			if (print_prompt("[in both] ")) {
				return -1;
			}
			kanji_selection = new_select;
		}
	}

	kdict.sort_by_score(kanji_selection);

	if (kanji_display_selection(10) == -1) {
		return -1;
	}
	return 0;
}

int
pt_link::canon_in::select_skip(char const *input) {
	list<int> select;
	skip_code sc(input);

	if (sc.packed() == 0) {
		log("@@@ input = \"");
		log_buf(input, strlen(input));
		log("\"\n");
		log("@@@ invalid skip code\n");
		return print_prompt('\a');
	}

	log("@@@ searching for skip %s\n", sc.unpacked());
	if (kdict.search_skip(sc, select) != 0) {
		return -1;
	}

	return merge_selection(select);
}

int
pt_link::canon_in::select_reading(char const *input, char cmd) {
	wide_char *roman = jis_to_wide_dup(input);
	if (roman == NULL) {
		lerror("jis_to_wide_dup");
		return -1;
	}
	size_t len = wide_len(roman) + 1;
	wide_char *kana = (wide_char *) malloc(len * sizeof(wide_char));
	if (kana == NULL) {
		free(roman);
		return -1;
	}
	wide_char const *tmp = roman;
	wide_roman_to_kana_str(kana, &len, &tmp, r2k_flags);
	assert(len > 0 && kana[len - 1] == '\0');
	free(roman);

	if (cmd == 'o') {
		wide_to_katakana_str(kana);
	} else if (cmd == 'k') {
		wide_to_hiragana_str(kana);
	}

	{
		char *tmp = wide_to_jis_dup(kana);
		log("@@@ input = '%s'; kana = '%s'", input, tmp);
		free(tmp);
	}

	list<int> select;
	int ret;

	if (cmd == 'n') {
		ret = kdict.search_name_reading(kana, select);
	} else {
		ret = kdict.search_reading(kana, select, cmd == 'b');
	}
	free(kana);

	if (ret != 0) {
		lerror("search_reading");
		return ret;
	}

	return merge_selection(select);
}

int
pt_link::canon_in::process_kanji_input(char c) {
	assert(kinput != NULL);
	int r = kinput->process_canon_input(c);
	if (r == -1) {
		lerror("process_kanji_input: process_canon_input");
		return r;
	}

	char *input = fake_input.get_input();
	if (input == NULL) {
		return 0;
	}

	if (use_status_line) {
		kinput->status_line_prompt = NULL;
		kinput->sl_msg_offset = kinput->sl_msg_pos = 0;
#if 0
		if (update_status_line() != 0) {
			return -1;
		}
#endif
	}

	char *nl = strchr(input, '\n');
	if (nl != NULL) {
		*nl = '\0';
	}

	int ret = 0;
	int n = atoi(input);

	switch(kanji_command) {
	case 'p':
		ret = select_skip(input);
		break;

	case 'r':
		if (n < 1 || n > 214) {
			ret = print_prompt("\aInvalid Radical #.\n");
			break;
		}
		ret = merge_selection(kdict.search_radical(n));
		break;

	case 's':
		if (n < 1 || n > 1000) {
			ret = print_prompt("\aInvalid Stroke Count.\n");
			break;
		}
		ret = merge_selection(kdict.search_stroke(n));
		break;

	case 'z':
		free(input);
		return substitute_kanji_input(n);

	case 'b':
        case 'k':
	case 'o':
	case 'n':
		ret = select_reading(input, kanji_command);
		break;

	default:
		assert(0);
	}

	if (set_prompt_mode() != 0) {
		ret = -1;
	}

	free(input);
	return ret;
}

int
pt_link::canon_in::substitute_kanji_input(int n) {
	if (n < 0 || n >= kanji_selection.length()) {
		if (print_prompt("\aInvalid Selection.\n") == -1) {
			return -1;
		}
		return set_prompt_mode();
	}

	wide_char wbuf[2];
	
	wbuf[0] = kdict[kanji_selection[n]];
	wbuf[1] = '\0';

	if (set_normal_mode() == -1) {
		return -1;
	}

	if (!use_status_line && process_reprint() != 0) {
		return -1;
	}

	return substitute_input(canon_pos, wbuf, 1);
}

int
pt_link::canon_in::process_kanji_command(char c) {
	switch(c) {
	case 'p':
		return start_kanji_input("SKIP code: ", c);

	case 's':
		return start_kanji_input("Strokes: ", c);

	case 'r':
		return start_kanji_input("Radical #: ", c);

	case 'o':
		return start_kanji_input("On Reading: ", c);

	case 'k':
		return start_kanji_input("Kun Reading: ", c);

	case 'b':
		return start_kanji_input("On/Kun Reading: ", c);

	case 'n':
		return start_kanji_input("Name Reading: ", c);

	case 'z':
		return start_kanji_input("Select #: ", c);

	case 'c':
		sl_msg_offset = sl_msg_pos = 0;

		if (print_prompt("Selection Cleared.\n") != 0) {
			return -1;
		}
		kanji_selection = NULL;
		return set_prompt_mode();

	case 'q':
		sl_msg_offset = sl_msg_pos = 0;

		if (print_prompt("Aborted") != 0
		    || flush_prompt() != 0) {
			return -1;
		}
		return set_normal_mode();

	case 'h':
	case '?':
		sl_msg_offset = sl_msg_pos = 0;
		char const *msg;
		if (use_status_line) {
			msg = "Q)uit"
				" B)oth"
				" C)lear"
				" K)un"
				" O)n"
				" N)ame"
				" SKI(P)"
				" R)ad."
				" S)troke"
				" Z)Select";
		} else { 
			msg = "\n[a] Show All;"
				" [b] On/Kun Reading;"
				" [c] Clear;"
				" [k] Kun Reading;"
				" [n] Name Reading;"
				" [o] On Reading;"
				" [p] SKIP code;"
				" [r] Radical #;"
				" [s] Strokes;"
				" [z] Select;"
				" [0-9] Select;"
				" [q] Quit\n";
		}

		if (print_prompt(msg) != 0 || flush_prompt() != 0) {
			return -1;
		}
		return set_prompt_mode();

	case 'a':
		sl_msg_offset = sl_msg_pos = 0;

		if (!use_status_line && print_prompt("Show All\n") != 0) {
			return -1;
		}
		if (kanji_display_selection() != 0) {
			return -1;
		}
		return set_prompt_mode();

	case ' ':
		if (!use_status_line) {
			break;
		}
		sl_msg_offset += MESSAGE_SCROLL;
		if (sl_msg_offset + MESSAGE_LENGTH > sl_msg_pos) {
			sl_msg_offset = sl_msg_pos - MESSAGE_LENGTH;
			if (sl_msg_offset < 0) {
				sl_msg_offset = 0;
			}
		}
		return set_prompt_mode();

	case '\b':
	case '\177':
		if (!use_status_line) {
			break;
		}
		if (sl_msg_offset > MESSAGE_SCROLL) {
			sl_msg_offset -= MESSAGE_SCROLL;
		} else {
			sl_msg_offset = 0;
		}
		return set_prompt_mode();
	}

	sl_msg_offset = sl_msg_pos = 0;

	if (isdigit(c)) {
		if (print_prompt("Select #: ") != 0
		    || print_prompt(c) != 0
		    || print_prompt('\n') != 0) {
			return -1;
		}
		int n = c - '0';
		return substitute_kanji_input(n);	    
	}

	if (print_prompt("\aUnknown Kanji Selection Command."
			     " ([?] for help)\n") == -1) {
		return -1;
	}
	return set_prompt_mode();
}
	
int
pt_link::canon_in::process_canon_input(char c) {
	slave_termios.c_lflag &= ~FLUSHO;

	/*
	 * Ignore all the bytes of a multibyte character that won't
	 * fit into the buffer.
	 */
	if (ignore_bytes > 0) {
		if (euc_is_mb(c)) {
			ignore_bytes--;
			return 0;
		}
		ignore_bytes = 0;
	}
			
	switch(command_mode) {
	case NONE:
		break;

	case COMMAND:
		if (c == slave_termios.c_cc[VSWTCH]) {
			lnext = 1;
			command_mode = NONE;
			break;
		}
		return process_command(c);

	case KANJI:
		return process_kanji_command(c);

	case KANJI_INPUT:
		return process_kanji_input(c);
	}

	if ((slave_termios.c_lflag & PENDIN) == PENDIN) {
		char c = slave_termios.c_cc[VREPRINT];
		int r;
		if (c != _POSIX_VDISABLE) {
			r = process_echo(c);
			if (r != 0) {
				return r;
			}
		}
		r = process_reprint();
		if (r != 0) {
			return r;
		}
	}

	int translate = 1;
	int vchar = NCCS;
	if (lnext) {
		lnext = 0;
		translate = 0;
	} else if (c != _POSIX_VDISABLE) {
		for (vchar = 0; vchar < NCCS; vchar++) {
			if (!icanon && (vchar == VMIN || vchar == VTIME)) {
				continue;
			}
			if (c == slave_termios.c_cc[vchar]) {
				break;
			}
		}
	}

	const int NL = NCCS + 1;
	if (vchar == NCCS && c == '\n') {
		vchar = NL;
	}

	switch(vchar) {
	case VINTR:
		if (slave_termios.c_lflag & ISIG) {
			return process_interrupt(c, SIGINT);
		}
		break;

	case VQUIT:
		if (slave_termios.c_lflag & ISIG) {
			return process_interrupt(c, SIGQUIT);
		}
		break;

	case VERASE:
		if (!backslash && icanon) {
			return process_erase(slave_termios.c_lflag & ECHOE);
		}
		break;

	case VWERASE:
		if (!backslash && icanon) {
			int echoe = (slave_termios.c_lflag & ECHOE) == ECHOE;
			while(canon_pos > 0 
			      && isspace(buf[canon_pos - 1])) {
				int r =	process_erase(echoe);
				if (r != 0) {
					return r;
				}
			}
			while(canon_pos > 0 
			      && !isspace(buf[canon_pos - 1])) {
				int r =	process_erase(echoe);
				if (r != 0) {
					return r;
				}
			}
			return 0;
		}
		break;

	case VKILL:
		if (!backslash && icanon) {
			if ((slave_termios.c_lflag & (ECHO | ECHOKE | IEXTEN))
 			    == (ECHO | ECHOKE | IEXTEN)) {
				int r =	process_erase(1, 0);
				if (r != 0) {
					return r;
				}
				return 0;
			}
			translation_pos = canon_pos = 0;
			update_status_line();
			if ((slave_termios.c_lflag & (ECHO | ECHOK)) 
			    == (ECHO | ECHOK)) {
				return process_echo('\n');
			}
			return 0;
		}
		break;

	case VREPRINT:
		if ((slave_termios.c_lflag & (ICANON | ECHO | IEXTEN))
		    == (ICANON | ECHO | IEXTEN)) {
			int r = process_echo(c);
			if (r != 0) {
				return r;
			}
			r = update_status_line(1);
			if (r != 0) {
				return r;
			}
			return process_reprint();
		}
		break;


	case VEOF:
		if (icanon) {
			int r = process_echo(c);
			if (r != 0) {
				return r;
			}
			return process_break();
		}
		break;

	case NL:
	case VEOL:
	case VEOL2:
	{
		if (icanon) {
			int r = process_character(c, 0);
			if (r != 0) {
				return r;
			}
			return process_break();
		}
		break;
	}

	case VSWTCH:
		log("@@@ got command char\n");
		if (!use_status_line) {
			slave_termios.c_lflag |= PENDIN;
		}
		command_mode = COMMAND;
		if (use_status_line) {
			return update_status_line();
		} else {
			return process_echo(c);
		}
		break;

	case VSUSP:
		return process_interrupt(c, SIGTSTP);

	case VDSUSP:
		return process_interrupt(c, SIGTSTP); /* ??? */

#if 0
	case VSTOP:
		if (slave_termios.c_iflag & IXON) {
			tcflow(ttyout, TCOOFF);
			stopped = 1;
			return 0;
		}
		break;

	case VSTRT:
		if (slave_termios.c_iflag & IXON) {
			tcflow(ttyout, TCOON);
			stopped = 0;
			return 0;
		}
		break;
#endif

	case VDISCARD:
		if ((slave_termios.c_lflag & (ICANON | IEXTEN))
		    == (ICANON | IEXTEN)) {
			flush_output(); /* ??? */
			slave_termios.c_lflag |= FLUSHO;
			return process_echo(c);
		}
		break;

	case VLNEXT:
		if ((slave_termios.c_lflag & (ICANON | IEXTEN))
		    == (ICANON | IEXTEN)) {
			lnext = 1;
			if (echo) {
				return write_echo("^\b");
			}
			return 0;
		}
		break;
	}

	if (c == '\\') {
		backslash = 1;
	} else {
		backslash = 0;
	}

	return process_character(c, translate);
}


int
pt_link::canon_in::process_setattr(int cmd, struct termios *tio) {
	if (tio->c_cc[VSWTCH] == _POSIX_VDISABLE) {
		char swtch = slave_termios.c_cc[VSWTCH];
		slave_termios = *tio;
		slave_termios.c_cc[VSWTCH] = swtch;
	} else {
		slave_termios = *tio;
	}
		
	master_termios.c_cflag = (master_termios.c_cflag & ~HUPCL)
	                         | (tio->c_cflag & HUPCL);
	master_termios.c_iflag = tio->c_iflag;
	if (command_mode == KANJI || !recursive_edit) {
		master_termios.c_oflag = tio->c_oflag;
	}

	echo = tio->c_lflag & ECHO;

	if (tio->c_lflag & ICANON) {
		icanon = 1;
		master_termios.c_cc[VMIN] = 1;
		master_termios.c_cc[VTIME] = 0;
	} else {
		icanon = 0;
		master_termios.c_cc[VMIN] = tio->c_cc[VMIN];
		master_termios.c_cc[VTIME] = tio->c_cc[VMIN];
	}

	log("@@@ setattr i: %07lo o: %07lo l: %07lo c: %07lo\n",
	    master_termios.c_iflag, master_termios.c_oflag,
	    master_termios.c_lflag, master_termios.c_cflag);

	if (pending_output.setattr(cmd, &master_termios) == -1) {
		lerror("ioctl setattr");
		return -1;
	}

	return 0;
}

int
pt_link::canon_in::process_termio(int cmd, struct termio *otio) {
	struct termios tio = slave_termios;

	tio.c_iflag = otio->c_iflag;
	tio.c_oflag = otio->c_oflag;
	tio.c_cflag = otio->c_cflag;
	tio.c_lflag = otio->c_lflag;
	
	int i;
	for (i = 0; i < NCC; i++) {
		tio.c_cc[i] = otio->c_cc[i];
	}

	switch(cmd) {
	case TCSETA:
		return process_setattr(TCSETS, &tio);

	case TCSETAW:
		return process_setattr(TCSETSW, &tio);

	case TCSETAF:
		return process_setattr(TCSETSF, &tio);
	}

	abort();
}

/* Local variables: */
/* mode: c++ */
/* End: */
